import csv
import threading
import traceback

from flask import g, redirect

from sqlalchemy.orm import joinedload

from chellow.dloads import open_file
from chellow.models import Llfc, Session, User
from chellow.utils import hh_format


def content(user_id):
    f = writer = None
    try:
        with Session() as sess:
            user = User.get_by_id(sess, user_id)
            f = open_file("llfcs.csv", user, mode="w", newline="")
            writer = csv.writer(f, lineterminator="\n")
            writer.writerow(
                (
                    "Chellow Id",
                    "DNO Code",
                    "Code",
                    "Description",
                    "Voltage Level",
                    "Is Substation?",
                    "Is Import?",
                    "Valid From",
                    "Valid To",
                )
            )

            for llfc in (
                sess.query(Llfc)
                .order_by(Llfc.id)
                .options(joinedload(Llfc.dno), joinedload(Llfc.voltage_level))
            ):
                writer.writerow(
                    (
                        str(llfc.id),
                        llfc.dno.dno_code,
                        llfc.code,
                        llfc.description,
                        llfc.voltage_level.code,
                        llfc.is_substation,
                        llfc.is_import,
                        hh_format(llfc.valid_from),
                        hh_format(llfc.valid_to),
                    )
                )
    except BaseException:
        msg = traceback.format_exc()
        print(msg)
        if writer is not None:
            writer.writerow([msg])
    finally:
        if f is not None:
            f.close()


def do_get(sess):
    args = (g.user.id,)
    threading.Thread(target=content, args=args).start()
    return redirect("/downloads", 303)



from collections import defaultdict
from datetime import datetime as Datetime
from decimal import Decimal

from openpyxl import Workbook


from chellow.gas.bill_parser_bgs_xlsx import _make_raw_bills, _parse_row
from chellow.utils import ct_datetime, to_utc, utc_datetime


def test_parse_row():
    wb = Workbook()
    ws = wb.active

    rows = [
        [
            "Customer",
            "Account",
            "MPRN",
            "Bill",
            "SupplyAddress",
            "BillDate",
            "BillingPeriod",
            "kWhs",
            "ChargeType",
            "ChargePeriodFrom",
            "ChargePeriodEnd",
            "Quantity",
            "QUnit",
            "Charge",
            "CUnit",
            "Total",
        ],
        [
            "Zorin Industries",
            69389283,
            86762618,
            584773269,
            "Extinct Volcano",
            Datetime(2025, 6, 13),
            Datetime(2025, 5, 1),
            0,
            "Gas",
            Datetime(2025, 5, 1),
            Datetime(2025, 5, 31),
            0,
            "kWh",
            0.5311,
            "p/kWh",
            0.00,
        ],
    ]

    for row, row_values in enumerate(rows, start=1):
        for column, value in enumerate(row_values, start=1):
            ws.cell(row=row, column=column, value=value)

    row = 2
    title_row = tuple(ws.rows)[0]
    bills = {}
    _parse_row(bills, ws, row, title_row)


def test_make_raw_bills():
    wb = Workbook()
    ws = wb.active

    rows = [
        [
            "Customer",
            "Account",
            "MPRN",
            "Bill",
            "SupplyAddress",
            "BillDate",
            "BillingPeriod",
            "kWhs",
            "ChargeType",
            "ChargePeriodFrom",
            "ChargePeriodEnd",
            "Quantity",
            "QUnit",
            "Charge",
            "CUnit",
            "Total",
        ],
        [
            "Zorin Industries",
            69389283,
            86762618,
            584773269,
            "Extinct Volcano",
            Datetime(2025, 6, 13),
            Datetime(2025, 5, 1),
            0,
            "Gas",
            Datetime(2025, 5, 1),
            Datetime(2025, 5, 31),
            0,
            "kWh",
            0.5311,
            "p/kWh",
            0.00,
        ],
    ]

    for row, row_vals in enumerate(rows, start=1):
        for column, value in enumerate(row_vals, start=1):
            ws.cell(row=row, column=column, value=value)

    bills = _make_raw_bills(ws)
    expected_bills = [
        {
            "bill_type_code": "N",
            "mprn": "86762618",
            "reference": "584773269",
            "account": 69389283,
            "issue_date": utc_datetime(2025, 6, 12, 23, 0),
            "start_date": utc_datetime(2025, 4, 30, 23, 0),
            "finish_date": utc_datetime(2025, 5, 31, 22, 30),
            "kwh": Decimal("0"),
            "breakdown": {
                "admin_variable_gbp": Decimal("0.0"),
                "admin_variable_kwh": Decimal("0"),
                "admin_variable_rate": {Decimal("0.005311")},
            },
            "net_gbp": Decimal("0.00"),
            "vat_gbp": Decimal("0.00"),
            "gross_gbp": Decimal("0.00"),
            "raw_lines": [
                [
                    "Zorin Industries",
                    69389283,
                    86762618,
                    584773269,
                    "Extinct Volcano",
                    Datetime(2025, 6, 13, 0, 0),
                    Datetime(2025, 5, 1, 0, 0),
                    0,
                    "Gas",
                    Datetime(2025, 5, 1, 0, 0),
                    Datetime(2025, 5, 31, 0, 0),
                    0,
                    "kWh",
                    0.5311,
                    "p/kWh",
                    0.0,
                ]
            ],
            "reads": [],
        }
    ]
    assert bills == expected_bills


def test_make_raw_bills_part_month():
    wb = Workbook()
    ws = wb.active

    rows = [
        [
            "Customer",
            "Account",
            "MPRN",
            "Bill",
            "SupplyAddress",
            "BillDate",
            "BillingPeriod",
            "kWhs",
            "ChargeType",
            "ChargePeriodFrom",
            "ChargePeriodEnd",
            "Quantity",
            "QUnit",
            "Charge",
            "CUnit",
            "Total",
        ],
        [
            "Zorin Industries",
            69389283,
            86762618,
            584773269,
            "Extinct Volcano",
            Datetime(2025, 6, 13),
            Datetime(2025, 5, 1),
            0,
            "Gas",
            Datetime(2025, 5, 1),
            Datetime(2025, 5, 15),
            0,
            "kWh",
            0.5311,
            "p/kWh",
            0.00,
        ],
    ]

    for row, row_vals in enumerate(rows, start=1):
        for column, value in enumerate(row_vals, start=1):
            ws.cell(row=row, column=column, value=value)

    bills = _make_raw_bills(ws)
    expected_bills = [
        {
            "bill_type_code": "N",
            "mprn": "86762618",
            "reference": "584773269",
            "account": 69389283,
            "issue_date": utc_datetime(2025, 6, 12, 23, 0),
            "start_date": to_utc(ct_datetime(2025, 5, 1, 0, 0)),
            "finish_date": to_utc(ct_datetime(2025, 5, 15, 23, 30)),
            "kwh": Decimal("0"),
            "breakdown": defaultdict(
                int,
                {
                    "admin_variable_gbp": Decimal("0.0"),
                    "admin_variable_kwh": Decimal("0"),
                    "admin_variable_rate": {Decimal("0.005311")},
                },
            ),
            "net_gbp": Decimal("0.00"),
            "vat_gbp": Decimal("0.00"),
            "gross_gbp": Decimal("0.00"),
            "raw_lines": [
                [
                    "Zorin Industries",
                    69389283,
                    86762618,
                    584773269,
                    "Extinct Volcano",
                    Datetime(2025, 6, 13, 0, 0),
                    Datetime(2025, 5, 1, 0, 0),
                    0,
                    "Gas",
                    Datetime(2025, 5, 1, 0, 0),
                    Datetime(2025, 5, 15, 0, 0),
                    0,
                    "kWh",
                    0.5311,
                    "p/kWh",
                    0.0,
                ],
            ],
            "reads": [],
        }
    ]
    assert bills == expected_bills


def test_make_raw_bills_credit():
    wb = Workbook()
    ws = wb.active

    rows = [
        [
            "Customer",
            "Account",
            "MPRN",
            "Bill",
            "SupplyAddress",
            "BillDate",
            "BillingPeriod",
            "kWhs",
            "ChargeType",
            "ChargePeriodFrom",
            "ChargePeriodEnd",
            "Quantity",
            "QUnit",
            "Charge",
            "CUnit",
            "Total",
        ],
        [
            "Zorin Industries",
            69389283,
            86762618,
            "CR584773269",
            "Extinct Volcano",
            Datetime(2025, 6, 13),
            Datetime(2025, 5, 1),
            0,
            "Gas",
            Datetime(2025, 5, 1),
            Datetime(2025, 5, 31),
            10,
            "kWh",
            0.5311,
            "p/kWh",
            -5.00,
        ],
    ]

    for row, row_vals in enumerate(rows, start=1):
        for column, value in enumerate(row_vals, start=1):
            ws.cell(row=row, column=column, value=value)

    bills = _make_raw_bills(ws)
    expected_bills = [
        {
            "bill_type_code": "W",
            "mprn": "86762618",
            "reference": "CR584773269",
            "account": 69389283,
            "issue_date": utc_datetime(2025, 6, 12, 23, 0),
            "start_date": utc_datetime(2025, 4, 30, 23, 0),
            "finish_date": utc_datetime(2025, 5, 31, 22, 30),
            "kwh": Decimal("0"),
            "breakdown": {
                "admin_variable_gbp": Decimal("-5.00"),
                "admin_variable_kwh": Decimal("-10"),
                "admin_variable_rate": {Decimal("0.005311")},
            },
            "net_gbp": Decimal("0.00"),
            "vat_gbp": Decimal("0.00"),
            "gross_gbp": Decimal("0.00"),
            "raw_lines": [
                [
                    "Zorin Industries",
                    69389283,
                    86762618,
                    "CR584773269",
                    "Extinct Volcano",
                    Datetime(2025, 6, 13, 0, 0),
                    Datetime(2025, 5, 1, 0, 0),
                    0,
                    "Gas",
                    Datetime(2025, 5, 1, 0, 0),
                    Datetime(2025, 5, 31, 0, 0),
                    10,
                    "kWh",
                    0.5311,
                    "p/kWh",
                    -5.0,
                ]
            ],
            "reads": [],
        }
    ]
    assert bills == expected_bills


def test_make_raw_bills_multi_date():
    wb = Workbook()
    ws = wb.active

    rows = [
        [
            "Customer",
            "Account",
            "MPRN",
            "Bill",
            "SupplyAddress",
            "BillDate",
            "BillingPeriod",
            "kWhs",
            "ChargeType",
            "ChargePeriodFrom",
            "ChargePeriodEnd",
            "Quantity",
            "QUnit",
            "Charge",
            "CUnit",
            "Total",
        ],
        [
            "Zorin Industries",
            69389283,
            86762618,
            584773269,
            "Extinct Volcano",
            Datetime(2025, 6, 13),
            Datetime(2025, 5, 1),
            0,
            "Gas",
            Datetime(2025, 5, 1),
            Datetime(2025, 5, 15),
            0,
            "kWh",
            0.5311,
            "p/kWh",
            0.00,
        ],
        [
            "Zorin Industries",
            69389283,
            86762618,
            584773269,
            "Extinct Volcano",
            Datetime(2025, 6, 13),
            Datetime(2025, 5, 1),
            0,
            "Gas",
            Datetime(2025, 5, 16),
            Datetime(2025, 5, 31),
            0,
            "kWh",
            0.5311,
            "p/kWh",
            0.00,
        ],
    ]

    for row, row_vals in enumerate(rows, start=1):
        for column, value in enumerate(row_vals, start=1):
            ws.cell(row=row, column=column, value=value)

    bills = _make_raw_bills(ws)
    expected_bills = [
        {
            "bill_type_code": "N",
            "mprn": "86762618",
            "reference": "584773269",
            "account": 69389283,
            "issue_date": utc_datetime(2025, 6, 12, 23, 0),
            "start_date": to_utc(ct_datetime(2025, 5, 1, 0, 0)),
            "finish_date": to_utc(ct_datetime(2025, 5, 31, 23, 30)),
            "kwh": Decimal("0"),
            "breakdown": defaultdict(
                int,
                {
                    "admin_variable_gbp": Decimal("0.0"),
                    "admin_variable_kwh": Decimal("0"),
                    "admin_variable_rate": {Decimal("0.005311")},
                },
            ),
            "net_gbp": Decimal("0.00"),
            "vat_gbp": Decimal("0.00"),
            "gross_gbp": Decimal("0.00"),
            "raw_lines": [
                [
                    "Zorin Industries",
                    69389283,
                    86762618,
                    584773269,
                    "Extinct Volcano",
                    Datetime(2025, 6, 13, 0, 0),
                    Datetime(2025, 5, 1, 0, 0),
                    0,
                    "Gas",
                    Datetime(2025, 5, 1, 0, 0),
                    Datetime(2025, 5, 15, 0, 0),
                    0,
                    "kWh",
                    0.5311,
                    "p/kWh",
                    0.0,
                ],
                [
                    "Zorin Industries",
                    69389283,
                    86762618,
                    584773269,
                    "Extinct Volcano",
                    Datetime(2025, 6, 13, 0, 0),
                    Datetime(2025, 5, 1, 0, 0),
                    0,
                    "Gas",
                    Datetime(2025, 5, 16, 0, 0),
                    Datetime(2025, 5, 31, 0, 0),
                    0,
                    "kWh",
                    0.5311,
                    "p/kWh",
                    0.0,
                ],
            ],
            "reads": [],
        }
    ]
    assert bills == expected_bills


def test_make_raw_bills_VAT():
    wb = Workbook()
    ws = wb.active

    rows = [
        [
            "Customer",
            "Account",
            "MPRN",
            "Bill",
            "SupplyAddress",
            "BillDate",
            "BillingPeriod",
            "kWhs",
            "ChargeType",
            "ChargePeriodFrom",
            "ChargePeriodEnd",
            "Quantity",
            "QUnit",
            "Charge",
            "CUnit",
            "Total",
        ],
        [
            "Zorin Industries",
            69389283,
            86762618,
            584773269,
            "Extinct Volcano",
            Datetime(2025, 6, 13),
            Datetime(2025, 5, 1),
            0,
            "VAT",
            None,
            None,
            0,
            "kWh",
            0.5311,
            "p/kWh",
            0.00,
        ],
        [
            "Zorin Industries",
            69389283,
            86762618,
            584773269,
            "Extinct Volcano",
            Datetime(2025, 6, 13),
            Datetime(2025, 5, 1),
            0,
            "Gas",
            Datetime(2025, 5, 16),
            Datetime(2025, 5, 31),
            0,
            "kWh",
            0.5311,
            "p/kWh",
            0.00,
        ],
    ]

    for row, row_vals in enumerate(rows, start=1):
        for column, value in enumerate(row_vals, start=1):
            ws.cell(row=row, column=column, value=value)

    bills = _make_raw_bills(ws)
    expected_bills = [
        {
            "bill_type_code": "N",
            "mprn": "86762618",
            "reference": "584773269",
            "account": 69389283,
            "issue_date": utc_datetime(2025, 6, 12, 23, 0),
            "start_date": to_utc(ct_datetime(2025, 5, 16, 0, 0)),
            "finish_date": to_utc(ct_datetime(2025, 5, 31, 23, 30)),
            "kwh": Decimal("0"),
            "breakdown": defaultdict(
                int,
                {
                    "admin_variable_gbp": Decimal("0.0"),
                    "admin_variable_kwh": Decimal("0"),
                    "admin_variable_rate": {Decimal("0.005311")},
                    "vat_rate": {Decimal("0.005311")},
                },
            ),
            "net_gbp": Decimal("0.00"),
            "vat_gbp": Decimal("0.00"),
            "gross_gbp": Decimal("0.00"),
            "raw_lines": [
                [
                    "Zorin Industries",
                    69389283,
                    86762618,
                    584773269,
                    "Extinct Volcano",
                    Datetime(2025, 6, 13, 0, 0),
                    Datetime(2025, 5, 1, 0, 0),
                    0,
                    "VAT",
                    None,
                    None,
                    0,
                    "kWh",
                    0.5311,
                    "p/kWh",
                    0.0,
                ],
                [
                    "Zorin Industries",
                    69389283,
                    86762618,
                    584773269,
                    "Extinct Volcano",
                    Datetime(2025, 6, 13, 0, 0),
                    Datetime(2025, 5, 1, 0, 0),
                    0,
                    "Gas",
                    Datetime(2025, 5, 16, 0, 0),
                    Datetime(2025, 5, 31, 0, 0),
                    0,
                    "kWh",
                    0.5311,
                    "p/kWh",
                    0.0,
                ],
            ],
            "reads": [],
        }
    ]
    assert bills == expected_bills



import csv
import sys
import threading
import traceback
from collections import OrderedDict, defaultdict
from datetime import datetime as Datetime
from decimal import Decimal
from itertools import combinations
from numbers import Number

from flask import g, redirect, request

from sqlalchemy import null, or_, select, true
from sqlalchemy.orm import joinedload

from werkzeug.exceptions import BadRequest

import chellow.gas.engine
from chellow.dloads import open_file
from chellow.models import (
    GBatch,
    GBill,
    GContract,
    GEra,
    RSession,
    ReportRun,
    Site,
    SiteGEra,
    User,
)
from chellow.utils import csv_make_val, hh_max, hh_min, req_date, req_int, to_utc


def content(
    g_batch_id,
    g_bill_id,
    g_contract_id,
    start_date,
    finish_date,
    user_id,
    report_run_id,
):
    forecast_date = to_utc(Datetime.max)
    report_context = {}
    sess = tmp_file = None
    try:
        with RSession() as sess:
            user = User.get_by_id(sess, user_id)

            tmp_file = open_file("g_bill_check.csv", user, mode="w")
            csv_writer = csv.writer(tmp_file)
            g_bills = (
                select(GBill)
                .order_by(GBill.g_supply_id, GBill.reference)
                .options(
                    joinedload(GBill.g_supply),
                    joinedload(GBill.g_batch),
                )
            )
            if g_batch_id is not None:
                g_batch = GBatch.get_by_id(sess, g_batch_id)
                g_contract = g_batch.g_contract
                g_bills = g_bills.where(GBill.g_batch == g_batch)

            elif g_bill_id is not None:
                g_bill = GBill.get_by_id(sess, g_bill_id)
                g_contract = g_bill.g_batch.g_contract
                g_bills = g_bills.where(GBill.id == g_bill.id)

            elif g_contract_id is not None:
                g_contract = GContract.get_by_id(sess, g_contract_id)
                g_bills = g_bills.join(GBatch).where(
                    GBatch.g_contract == g_contract,
                    GBill.start_date <= finish_date,
                    GBill.finish_date >= start_date,
                )

            vbf = chellow.gas.engine.g_contract_func(
                report_context, g_contract, "virtual_bill"
            )
            if vbf is None:
                raise BadRequest(
                    f"The contract {g_contract.name} doesn't have a function "
                    f"virtual_bill."
                )

            header_titles = [
                "batch",
                "bill_reference",
                "bill_type",
                "bill_start_date",
                "bill_finish_date",
                "mprn",
                "supply_name",
                "site_code",
                "site_name",
                "covered_start",
                "covered_finish",
                "covered_bill_ids",
            ]
            bill_titles = chellow.gas.engine.g_contract_func(
                report_context, g_contract, "virtual_bill_titles"
            )()

            titles = header_titles[:]
            for title in bill_titles:
                for prefix in ("covered_", "virtual_"):
                    titles.append(prefix + title)
                if title.endswith("_gbp"):
                    titles.append("difference_" + title)

            csv_writer.writerow(titles)

            g_bill_map = defaultdict(set, {})
            for b in sess.execute(g_bills).scalars():
                g_bill_map[b.g_supply.id].add(b.id)

            for g_supply_id, g_bill_ids in g_bill_map.items():
                while len(g_bill_ids) > 0:
                    _process_g_bill_ids(
                        sess,
                        report_context,
                        g_contract,
                        g_bill_ids,
                        forecast_date,
                        bill_titles,
                        vbf,
                        titles,
                        csv_writer,
                        report_run_id,
                    )
        ReportRun.w_update(report_run_id, "finished")
    except BadRequest as e:
        tmp_file.write(f"Problem: {e.description}")
        ReportRun.w_update(report_run_id, "problem")
    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg + "\n")
        tmp_file.write(f"Problem {msg}")
        ReportRun.w_update(report_run_id, "interrupted")
    finally:
        tmp_file.close()


def _process_g_bill_ids(
    sess,
    report_context,
    g_contract,
    g_bill_ids,
    forecast_date,
    bill_titles,
    vbf,
    titles,
    csv_writer,
    report_run_id,
):
    g_bill_id = list(sorted(g_bill_ids))[0]
    g_bill_ids.remove(g_bill_id)
    g_bill = sess.query(GBill).filter(GBill.id == g_bill_id).one()
    g_supply = g_bill.g_supply
    vals = {
        "covered_vat_gbp": Decimal("0.00"),
        "covered_net_gbp": Decimal("0.00"),
        "covered_gross_gbp": Decimal("0.00"),
        "covered_kwh": Decimal(0),
        "covered_start": g_bill.start_date,
        "covered_finish": g_bill.finish_date,
        "covered_bill_ids": [],
        "virtual_problem": "",
    }
    site_id = None
    read_dict = defaultdict(set)
    for g_read in g_bill.g_reads:
        if not all(
            g_read.msn == era.msn
            for era in g_supply.find_g_eras(sess, g_read.prev_date, g_read.pres_date)
        ):
            vals["virtual_problem"] += (
                f"The MSN {g_read.msn} of the register read {g_read.id} doesn't match "
                f"the MSN of all the relevant eras."
            )

        for dt, typ in [
            (g_read.pres_date, g_read.pres_type),
            (g_read.prev_date, g_read.prev_type),
        ]:
            typ_set = read_dict[str(dt) + "-" + g_read.msn]
            typ_set.add(typ)
            if len(typ_set) > 1:
                vals[
                    "virtual_problem"
                ] += f" Reads taken on {dt} have differing read types."

    covered_primary_bill = None
    enlarged = True

    while enlarged:
        enlarged = False
        covered_bills = OrderedDict(
            (b.id, b)
            for b in sess.query(GBill)
            .filter(
                GBill.g_supply == g_supply,
                GBill.start_date <= vals["covered_finish"],
                GBill.finish_date >= vals["covered_start"],
            )
            .order_by(GBill.issue_date.desc(), GBill.start_date)
        )

        num_covered = None
        while num_covered != len(covered_bills):
            num_covered = len(covered_bills)
            for a, b in combinations(tuple(covered_bills.values()), 2):
                if all(
                    (
                        a.start_date == b.start_date,
                        a.finish_date == b.finish_date,
                        a.kwh == -1 * b.kwh,
                        a.net == -1 * b.net,
                        a.vat == -1 * b.vat,
                        a.gross == -1 * b.gross,
                    )
                ):
                    for gb_id in a.id, b.id:
                        del covered_bills[gb_id]
                        if gb_id in g_bill_ids:
                            g_bill_ids.remove(gb_id)
                    break

        for covered_bill in covered_bills.values():
            if covered_primary_bill is None and len(covered_bill.g_reads) > 0:
                covered_primary_bill = covered_bill
            if covered_bill.start_date < vals["covered_start"]:
                vals["covered_start"] = covered_bill.start_date
                enlarged = True
                break
            if covered_bill.finish_date > vals["covered_finish"]:
                vals["covered_finish"] = covered_bill.finish_date
                enlarged = True
                break

    if len(covered_bills) == 0:
        return

    for covered_bill in covered_bills.values():
        if covered_bill.id in g_bill_ids:
            g_bill_ids.remove(covered_bill.id)
        vals["covered_bill_ids"].append(covered_bill.id)
        bdown = covered_bill.make_breakdown()
        vals["covered_kwh"] += covered_bill.kwh
        vals["covered_net_gbp"] += covered_bill.net
        vals["covered_vat_gbp"] += covered_bill.vat
        vals["covered_gross_gbp"] += covered_bill.gross
        for title in bill_titles:
            k = "covered_" + title
            v = bdown.get(title)

            if v is not None:
                if isinstance(v, list):
                    if k not in vals:
                        vals[k] = set()
                    vals[k].update(set(v))
                else:
                    try:
                        vals[k] += v
                    except KeyError:
                        vals[k] = v
                    except TypeError:
                        raise BadRequest(
                            f"Problem with bill {g_bill.id} and key {k} and value {v} "
                            f"for existing {vals[k]}"
                        )

            if title in (
                "correction_factor",
                "calorific_value",
                "unit_code",
                "unit_factor",
            ):
                if k not in vals:
                    vals[k] = set()
                for g_read in covered_bill.g_reads:
                    if title in ("unit_code", "unit_factor"):
                        g_unit = g_read.g_unit
                        if title == "unit_code":
                            v = g_unit.code
                        else:
                            v = g_unit.factor
                    else:
                        v = getattr(g_read, title)
                    vals[k].add(v)
    elements = set()
    for g_era in sess.execute(
        select(GEra).where(
            GEra.g_supply == g_supply,
            GEra.start_date <= vals["covered_finish"],
            or_(GEra.finish_date == null(), GEra.finish_date >= vals["covered_start"]),
        )
    ).scalars():
        if g_era.g_contract != g_contract:
            vals[
                "virtual_problem"
            ] += "The contract of the bill is different from the contract of the era."
            continue
        site = (
            sess.query(Site)
            .join(SiteGEra)
            .filter(SiteGEra.is_physical == true(), SiteGEra.g_era == g_era)
            .one()
        )
        site_id = site.id
        vals["site_code"] = site.code
        vals["site_name"] = site.name

        chunk_start = hh_max(vals["covered_start"], g_era.start_date)
        chunk_finish = hh_min(vals["covered_finish"], g_era.finish_date)

        data_source = chellow.gas.engine.GDataSource(
            sess,
            chunk_start,
            chunk_finish,
            forecast_date,
            g_era,
            report_context,
            covered_primary_bill,
        )

        vbf(data_source)

        for k, v in data_source.bill.items():
            vk = "virtual_" + k
            try:
                if isinstance(v, set):
                    vals[vk].update(v)
                else:
                    vals[vk] += v
            except KeyError:
                vals[vk] = v
            except TypeError as detail:
                raise BadRequest(f"For key {vk} and value {v}. {detail}")

            if k.endswith("_gbp"):
                elements.add(k[:-4])

    if g_bill.id not in covered_bills.keys():
        g_bill = covered_bills[sorted(covered_bills.keys())[0]]

    vals["batch"] = g_bill.g_batch.reference
    vals["bill_reference"] = g_bill.reference
    vals["bill_type"] = g_bill.bill_type.code
    vals["bill_start_date"] = g_bill.start_date
    vals["bill_finish_date"] = g_bill.finish_date
    vals["mprn"] = g_supply.mprn
    vals["supply_name"] = g_supply.name

    for i, title in enumerate(titles):
        if title.startswith("difference_"):
            covered_val = float(vals.get(titles[i - 2], 0))
            virtual_val = float(vals.get(titles[i - 1], 0))
            vals[title] = covered_val - virtual_val

    csv_writer.writerow([csv_make_val(vals.get(k)) for k in titles])
    vals["g_bill_id"] = g_bill.id
    vals["g_batch_id"] = g_bill.g_batch.id
    vals["g_supply_id"] = g_supply.id
    vals["site_id"] = site_id
    for element in sorted(elements):
        for key in tuple(vals.keys()):
            if not key.endswith("_gbp"):
                covered_prefix = f"covered_{element}_"
                virtual_prefix = f"virtual_{element}_"
                if key.startswith(covered_prefix):
                    part_name = key[len(covered_prefix) :]
                elif key.startswith(virtual_prefix):
                    part_name = key[len(virtual_prefix) :]
                else:
                    continue
                virtual_part = vals.get(f"virtual_{element}_{part_name}", {0})
                covered_part = vals.get(f"covered_{element}_{part_name}", {0})
                if isinstance(virtual_part, set) and len(virtual_part) == 1:
                    virtual_part = float(next(iter(virtual_part)))
                if isinstance(covered_part, set) and len(covered_part) == 1:
                    covered_part = float(next(iter(covered_part)))

                if isinstance(virtual_part, Number) and isinstance(
                    covered_part, Number
                ):
                    diff = float(covered_part) - float(virtual_part)
                else:
                    diff = 0

                vals[f"difference_{element}_{part_name}"] = diff

    for suffix in (
        "kwh",
        "units_consumed",
    ):
        try:
            covered = float(vals[f"covered_{suffix}"])
            virtual = float(vals[f"virtual_{suffix}"])
            vals[f"difference_{suffix}"] = covered - virtual
        except KeyError:
            vals[f"difference_{suffix}"] = 0

    for prefix in (
        "correction_factor",
        "unit_code",
        "calorific_value",
    ):
        covered = vals.get(f"covered_{prefix}", [0])
        virtual = vals.get(f"virtual_{prefix}", [0])
        if len(covered) == 1 and len(virtual) == 1:
            if covered == virtual:
                diff = 0
            else:
                try:
                    diff = float(next(iter(covered))) - float(next(iter(virtual)))
                except ValueError:
                    diff = False
        else:
            diff = False
        vals[f"difference_{prefix}"] = diff

    ReportRun.w_insert_row(report_run_id, "", titles, vals, {"is_checked": False})


def do_get(sess):
    g_batch_id = g_bill_id = g_contract_id = start_date = finish_date = None

    if "g_batch_id" in request.values:
        g_batch_id = req_int("g_batch_id")
        g_batch = GBatch.get_by_id(g.sess, g_batch_id)
        run_g_contract_id = g_batch.g_contract.id
    elif "g_bill_id" in request.values:
        g_bill_id = req_int("g_bill_id")
        g_bill = GBill.get_by_id(g.sess, g_bill_id)
        run_g_contract_id = g_bill.g_batch.g_contract.id
    elif "g_contract_id" in request.values:
        g_contract_id = req_int("g_contract_id")
        start_date = req_date("start_date")
        finish_date = req_date("finish_date")
        run_g_contract_id = g_contract_id
    else:
        raise BadRequest(
            "The bill check needs a g_batch_id, g_bill_id or g_contract_id."
        )

    report_run = ReportRun.insert(
        sess,
        "g_bill_check",
        g.user,
        "g_bill_check",
        {"g_contract_id": run_g_contract_id},
    )
    sess.commit()

    args = (
        g_batch_id,
        g_bill_id,
        g_contract_id,
        start_date,
        finish_date,
        g.user.id,
        report_run.id,
    )
    threading.Thread(target=content, args=args).start()
    return redirect(f"/report_runs/{report_run.id}", 303)



from collections import defaultdict
from decimal import Decimal
from io import BytesIO

from chellow.e.bill_parsers.sse_edi import (
    Parser,
    _process_BTL,
    _process_CCD2,
    _process_CCD3,
    _process_CCD4,
    _process_MTR,
)
from chellow.utils import ct_datetime, to_utc, utc_datetime


def test_missing_mpan_core(mocker, sess):
    file_lines = [
        "MHD=2+UTLBIL:3'",
        "BTL=+18511+8480++80612'",
        "MTR=6'",
    ]
    f = BytesIO(b"\n".join(n.encode("utf8") for n in file_lines))
    parser = Parser(f)
    bills = parser.make_raw_bills()
    expected_bill = {
        "bill_type_code": None,
        "account": None,
        "mpan_core": None,
        "reference": None,
        "issue_date": None,
        "start_date": None,
        "finish_date": None,
        "kwh": Decimal("0"),
        "net": Decimal("185.11"),
        "vat": Decimal("84.80"),
        "gross": Decimal("806.12"),
        "breakdown": defaultdict(
            int,
            {
                "raw-lines": [
                    "MHD=2+UTLBIL:3'",
                    "BTL=+18511+8480++80612'",
                    "MTR=6'",
                ]
            },
        ),
        "elements": [],
        "reads": [],
    }
    assert bills == [expected_bill]


def test_parser_REF(mocker, sess):
    file_lines = [
        "REF='",
    ]
    f = BytesIO(b"\n".join(n.encode("utf8") for n in file_lines))
    parser = Parser(f)
    parser.make_raw_bills()


def test_process_CCD2(mocker, sess):
    headers = {
        "reads": [],
        "breakdown": defaultdict(int),
        "kwh": Decimal(0),
        "elements": [],
    }
    elements = {
        "TMOD": ["0001"],
        "MTNR": ["x"],
        "MLOC": ["228477911812004111222"],
        "PRDT": ["201001"],
        "PVDT": ["200901"],
        "PRRD": ["10", "00", "0", "00"],
        "ADJF": ["", "0"],
        "CONA": ["0", "KWH"],
        "NUCT": ["0"],
        "CSDT": ["200701"],
        "CEDT": ["200901"],
        "CPPU": ["0"],
        "CTOT": ["0"],
    }
    _process_CCD2(elements, headers)
    expected_headers = {
        "reads": [
            {
                "msn": "x",
                "mpan": "04 111 222 22 8477 9118 120",
                "coefficient": Decimal("0"),
                "units": "kWh",
                "tpr_code": "00001",
                "prev_date": utc_datetime(2020, 9, 1, 22, 30),
                "prev_value": Decimal("0"),
                "prev_type_code": "N",
                "pres_date": utc_datetime(2020, 10, 1, 22, 30),
                "pres_value": Decimal("10"),
                "pres_type_code": "N",
            }
        ],
        "elements": [
            {
                "name": "00001",
                "net": Decimal("0"),
                "start_date": to_utc(ct_datetime(2020, 7, 1)),
                "finish_date": to_utc(ct_datetime(2020, 9, 1, 23, 30)),
                "breakdown": {
                    "kwh": Decimal("0"),
                    "rate": {Decimal("0")},
                },
            }
        ],
        "kwh": Decimal("0"),
        "breakdown": {},
    }
    assert headers == expected_headers


def test_process_CCD3(mocker, sess):
    headers = {"reads": [], "elements": []}
    elements = {
        "TMOD": ["0001", "CQ1UUP", "", "O"],
        "NUCT": ["8760000", "1"],
        "CPPU": ["086380"],
        "CTOT": ["37150"],
        "CSDT": ["151022"],
        "CEDT": ["171123"],
    }
    _process_CCD3(elements, headers)
    expected_headers = {
        "reads": [],
        "elements": [
            {
                "breakdown": {
                    "kwh": Decimal("8760"),
                    "rate": {
                        Decimal("0.8638"),
                    },
                },
                "finish_date": to_utc(ct_datetime(2017, 11, 23, 23, 30)),
                "name": "00001",
                "net": Decimal("371.50"),
                "start_date": to_utc(ct_datetime(2015, 10, 22)),
            }
        ],
    }
    assert headers == expected_headers
    assert headers["elements"][0]["net"].as_tuple().exponent == -2


def test_process_CCD4(mocker, sess):
    headers = {"reads": [], "elements": []}
    elements = {
        "NDRP": ["241"],
        "TMOD": ["0001", "CQ1UUP", "", "O"],
        "NUCT": ["8760000", "1"],
        "CPPU": ["086380"],
        "CTOT": ["37150"],
        "CSDT": ["151022"],
        "CEDT": ["171123"],
    }
    _process_CCD4(elements, headers)
    expected_headers = {
        "reads": [],
        "elements": [
            {
                "breakdown": {
                    "days": Decimal("241"),
                },
                "finish_date": to_utc(ct_datetime(2017, 11, 23, 23, 30)),
                "name": "standing",
                "net": Decimal("371.50"),
                "start_date": to_utc(ct_datetime(2015, 10, 22)),
            }
        ],
    }
    assert headers == expected_headers
    assert headers["elements"][0]["net"].as_tuple().exponent == -2


def test_process_BTL(mocker, sess):
    issue_date = to_utc(ct_datetime(2020, 10, 1))
    start_date = to_utc(ct_datetime(2020, 9, 1))
    finish_date = to_utc(ct_datetime(2020, 9, 30, 23, 30))
    bill_type_code = "N"
    mpan_core = "22 8477 9118 129"
    account = "AC2"
    reference = "h98ge4kl"
    kwh = Decimal("23")
    net = "45.00"
    vat = "5.20"
    gross = "27.60"
    breakdown = {"raw-lines": []}

    headers = {
        "is_ebatch": True,
        "reads": [{"pres_type_code": "C"}],
        "message_type": "UTLBIL",
        "mpan_core": mpan_core,
        "bill_type_code": bill_type_code,
        "account": account,
        "reference": reference,
        "issue_date": issue_date,
        "start_date": start_date,
        "finish_date": finish_date,
        "breakdown": breakdown,
        "kwh": kwh,
        "elements": [],
    }
    elements = {"UVLT": ["4500"], "UTVA": ["520"], "TBTL": ["2760"]}
    bill = _process_BTL(elements, headers)
    expected_bill = {
        "bill_type_code": bill_type_code,
        "account": account,
        "mpan_core": mpan_core,
        "reference": reference,
        "issue_date": issue_date,
        "start_date": start_date,
        "finish_date": finish_date,
        "kwh": kwh,
        "net": Decimal(net),
        "vat": Decimal(vat),
        "gross": Decimal(gross),
        "breakdown": breakdown,
        "reads": [{"pres_type_code": "E"}],
        "elements": [],
    }
    assert bill == expected_bill
    assert bill["net"].as_tuple().exponent == -2
    assert bill["vat"].as_tuple().exponent == -2
    assert bill["gross"].as_tuple().exponent == -2


def test_process_MTR(mocker, sess):
    headers = {}
    elements = {}
    bill = _process_MTR(elements, headers)
    assert bill is None



from datetime import datetime as Datetime
from decimal import Decimal


from sqlalchemy import select

from chellow.e.lcc import api_records
from chellow.models import Contract, RateScript
from chellow.utils import ct_datetime, to_ct, to_utc


def hh(data_source):
    try:
        rab_cache = data_source.caches["rab"]
    except KeyError:
        rab_cache = data_source.caches["rab"] = {}

    for h in data_source.hh_data:
        try:
            h["rab"] = rab_cache[h["start-date"]]
        except KeyError:
            h_start = h["start-date"]
            rate_str = data_source.non_core_rate("rab_forecast_ilr_tra", h_start)[
                "record"
            ]["Interim_Levy_Rate_GBP_MWh"]
            if rate_str == "":
                base_rate_dec = Decimal("0")
            else:
                base_rate_dec = Decimal(rate_str) / Decimal(1000)

            base_rate = float(base_rate_dec)

            h["rab"] = rab_cache[h_start] = {
                "interim": base_rate,
            }


def lcc_import(sess, log, set_progress, s):
    import_forecast_ilr_tra(sess, log, set_progress, s)


def _parse_date(date_str):
    return to_utc(to_ct(Datetime.strptime(date_str[:10], "%Y-%m-%d")))


def import_forecast_ilr_tra(sess, log, set_progress, s):
    log("Starting to check for new LCC RAB Forecast ILR TRA")

    contract_name = "rab_forecast_ilr_tra"
    contract = Contract.find_non_core_by_name(sess, contract_name)
    if contract is None:
        contract = Contract.insert_non_core(
            sess, contract_name, "", {}, to_utc(ct_datetime(1996, 4, 1)), None, {}
        )

    for record in api_records(log, s, "1231fbb3-93ee-4a33-87a9-f15bb377346d"):
        period_start_str = record["Month"]
        if len(period_start_str) == 0:
            continue
        period_start = _parse_date(period_start_str)

        rs = sess.execute(
            select(RateScript).where(
                RateScript.contract == contract,
                RateScript.start_date == period_start,
            )
        ).scalar_one_or_none()
        if rs is None:
            rs = contract.insert_rate_script(sess, period_start, {})

        rs_script = rs.make_script()
        rs_script["record"] = record
        rs.update(rs_script)
        sess.commit()
    log("Finished LCC RAB Forecast ILR TRA")



import csv
import sys
import threading
import traceback

from flask import g, redirect

from sqlalchemy import or_
from sqlalchemy.sql.expression import null, true

import chellow.e.computer
from chellow.dloads import open_file
from chellow.e.computer import contract_func
from chellow.gas.engine import GDataSource
from chellow.models import GEra, GSupply, Session, Site, SiteGEra
from chellow.utils import csv_make_val, hh_format, hh_range, req_date, req_int


def content(g_supply_id, start_date, finish_date, user):
    caches = {}
    try:
        with Session() as sess:
            g_supply = GSupply.get_by_id(sess, g_supply_id)

            forecast_date = chellow.e.computer.forecast_date()

            prev_titles = None
            f = open_file(
                f"g_supply_virtual_bills_hh_{g_supply_id}.csv",
                user,
                mode="w",
                newline="",
            )
            w = csv.writer(f, lineterminator="\n")

            for hh_start in hh_range(caches, start_date, finish_date):
                g_era = (
                    sess.query(GEra)
                    .filter(
                        GEra.g_supply == g_supply,
                        GEra.start_date <= hh_start,
                        or_(GEra.finish_date == null(), GEra.finish_date >= hh_start),
                    )
                    .one()
                )

                site = (
                    sess.query(Site)
                    .join(SiteGEra)
                    .filter(SiteGEra.g_era == g_era, SiteGEra.is_physical == true())
                    .one()
                )

                ds = GDataSource(
                    sess, hh_start, hh_start, forecast_date, g_era, caches, None
                )

                titles = ["MPRN", "Site Code", "Site Name", "Account", "HH Start", ""]

                output_line = [
                    ds.mprn,
                    site.code,
                    site.name,
                    ds.account,
                    hh_format(ds.start_date),
                    "",
                ]

                contract = g_era.g_contract
                output_line.append("")
                contract_titles = contract_func(
                    caches, contract, "virtual_bill_titles"
                )()
                titles.append("")
                titles.extend(contract_titles)

                contract_func(caches, contract, "virtual_bill")(ds)
                bill = ds.bill
                for title in contract_titles:
                    output_line.append(csv_make_val(bill.get(title, "")))
                    if title in bill:
                        del bill[title]

                for k in sorted(bill.keys()):
                    output_line.extend([k, csv_make_val(bill[k])])

                if titles != prev_titles:
                    prev_titles = titles
                    w.writerow(titles)
                w.writerow(output_line)
    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg)
        w.writerow([msg])
    finally:
        if f is not None:
            f.close()


def do_get(sess):
    g_supply_id = req_int("g_supply_id")
    start_date = req_date("start")
    finish_date = req_date("finish")

    args = g_supply_id, start_date, finish_date, g.user
    threading.Thread(target=content, args=args).start()
    return redirect("/downloads", 303)



import datetime
import traceback
from datetime import timedelta as Timedelta

from sqlalchemy import select

from werkzeug.exceptions import BadRequest

import xlrd

from zish import loads

from chellow.models import Contract, RateScript
from chellow.utils import HH, ct_datetime, hh_format, to_ct, to_utc


def key_format(dt):
    return dt.strftime("%d %H:%M")


def hh(data_source):
    for h in data_source.hh_data:
        try:
            sbp, ssp = data_source.caches["system_price"][h["start-date"]]
        except KeyError:
            try:
                system_price_cache = data_source.caches["system_price"]
            except KeyError:
                system_price_cache = data_source.caches["system_price"] = {}

            h_start = h["start-date"]
            rates = data_source.non_core_rate("system_price", h_start)[
                "gbp_per_nbp_mwh"
            ]

            try:
                try:
                    rdict = rates[key_format(h_start)]
                except KeyError:
                    rdict = rates[key_format(h_start - Timedelta(days=3))]
                sbp = float(rdict["sbp"] / 1000)
                ssp = float(rdict["ssp"] / 1000)
                system_price_cache[h_start] = (sbp, ssp)
            except KeyError:
                raise BadRequest(
                    f"For the System Price rate script at {hh_format(h_start)} "
                    f"the rate cannot be found."
                )
            except TypeError:
                raise BadRequest(
                    f"For the System Price rate script at {hh_format(h_start)} "
                    f"the rate 'rates_gbp_per_mwh' has the problem: "
                    f"{traceback.format_exc()}"
                )

        h["sbp"] = sbp
        h["sbp-gbp"] = h["nbp-kwh"] * sbp

        h["ssp"] = ssp
        h["ssp-gbp"] = h["nbp-kwh"] * ssp


def elexon_import(sess, log, set_progress, s, scripting_key):
    log("Starting to check System Prices.")
    contract_name = "system_price"
    contract = Contract.find_non_core_by_name(sess, contract_name)
    if contract is None:
        contract = Contract.insert_non_core(
            sess,
            contract_name,
            "",
            {"enabled": True},
            to_utc(ct_datetime(1996, 4, 1)),
            None,
            {},
        )
        sess.commit()
    contract_props = contract.make_properties()

    if not contract_props.get("enabled", False):
        log(
            "The automatic importer is disabled. To enable it, edit "
            "the contract properties to set 'enabled' to True."
        )
        return

    for rscript in sess.scalars(
        select(RateScript)
        .where(RateScript.contract == contract)
        .order_by(RateScript.start_date.desc())
    ):
        ns = loads(rscript.script)
        rates = ns.get("gbp_per_nbp_mwh", {})
        if len(rates) == 0:
            fill_start = rscript.start_date
            break
        elif rates[key_format(rscript.finish_date)]["run"] == "DF":
            fill_start = rscript.finish_date + HH
            break
    params = {"key": scripting_key}
    url = "https://downloads.elexonportal.co.uk/file/download/BESTVIEWPRICES_FILE"

    log(
        f"Downloading from {url}?key={scripting_key} and extracting data from "
        f"{hh_format(fill_start)}"
    )

    sess.rollback()  # Avoid long-running transactions
    res = s.get(url, params=params)
    log(f"Received {res.status_code} {res.reason}")
    data = res.content
    book = xlrd.open_workbook(file_contents=data)
    sbp_sheet = book.sheet_by_index(1)
    ssp_sheet = book.sheet_by_index(2)

    sp_months = []
    sp_month = None
    for row_index in range(1, sbp_sheet.nrows):
        sbp_row = sbp_sheet.row(row_index)
        ssp_row = ssp_sheet.row(row_index)
        raw_date = datetime.datetime(
            *xlrd.xldate_as_tuple(sbp_row[0].value, book.datemode)
        )
        hh_date_ct = to_ct(raw_date)
        hh_date = to_utc(hh_date_ct)
        run_code = sbp_row[1].value
        for col_idx in range(2, 52):
            if hh_date >= fill_start:
                sbp_val = sbp_row[col_idx].value
                if sbp_val != "":
                    if hh_date.day == 1 and hh_date.hour == 0 and hh_date.minute == 0:
                        sp_month = {}
                        sp_months.append(sp_month)
                    ssp_val = ssp_row[col_idx].value
                    if sp_month is not None:
                        sp_month[hh_date] = {
                            "run": run_code,
                            "sbp": sbp_val,
                            "ssp": ssp_val,
                        }
            hh_date += HH
    log("Successfully extracted data.")
    last_date = sorted(sp_months[-1].keys())[-1]
    if last_date.month == (last_date + HH).month:
        del sp_months[-1]
    if "limit" in contract_props:
        sp_months = sp_months[0:1]
    for sp_month in sp_months:
        sorted_keys = sorted(sp_month.keys())
        month_start = sorted_keys[0]
        month_finish = sorted_keys[-1]
        rs = (
            sess.query(RateScript)
            .filter(
                RateScript.contract == contract,
                RateScript.start_date == month_start,
            )
            .first()
        )
        if rs is None:
            log(f"Adding a new rate script starting at {hh_format(month_start)}.")

            latest_rs = (
                sess.query(RateScript)
                .filter(RateScript.contract == contract)
                .order_by(RateScript.start_date.desc())
                .first()
            )

            contract.update_rate_script(
                sess,
                latest_rs,
                latest_rs.start_date,
                month_finish,
                loads(latest_rs.script),
            )
            rs = contract.insert_rate_script(sess, month_start, {})
            sess.flush()
        script = {
            "gbp_per_nbp_mwh": dict((key_format(k), v) for k, v in sp_month.items())
        }
        log(f"Updating rate script starting at {hh_format(month_start)}.")
        contract.update_rate_script(sess, rs, rs.start_date, rs.finish_date, script)
        sess.commit()



from datetime import datetime as Datetime
from decimal import Decimal, InvalidOperation

from dateutil.relativedelta import relativedelta

from openpyxl import load_workbook

from werkzeug.exceptions import BadRequest


from chellow.utils import parse_mpan_core, to_ct, to_utc


def get_ct_date(row, idx):
    cell = get_cell(row, idx)
    val = cell.value
    if isinstance(val, Datetime):
        dt = val
    elif isinstance(val, str):
        dt = Datetime.strptime(val, "dd/mm/yyyy")
    else:
        raise BadRequest(
            f"The value {val} at {cell.coordinate} is of type {type(val)}, but "
            f"expected a timestamp or string."
        )
    return to_ct(dt)


def get_cell(row, idx):
    try:
        return row[idx]
    except IndexError:
        raise BadRequest(
            f"For the row {row}, the index is {idx} which is beyond the end of the row."
        )


def get_str(row, idx):
    return get_cell(row, idx).value.strip()


def get_dec(row, idx):
    cell = get_cell(row, idx)
    try:
        return Decimal(str(cell.value))
    except InvalidOperation as e:
        raise BadRequest(f"Problem parsing the number at {cell.coordinate}. {e}")


def get_int(row, idx):
    return int(get_cell(row, idx).value)


def _process_row(issue_date, row):

    mpan_core = parse_mpan_core(str(get_int(row, 1)))
    comm = get_str(row, 2)

    settled_str = get_str(row, 3)
    if settled_str == "Settled":
        settlement_status = "settlement"
    else:
        settlement_status = "non_settlement"

    start_date_ct = get_ct_date(row, 5)
    finish_date_ct = get_ct_date(row, 6)
    days = (finish_date_ct - start_date_ct).days + 1

    start_date = to_utc(start_date_ct)
    finish_date = to_utc(finish_date_ct + relativedelta(hours=23, minutes=30))

    meter_rate = get_dec(row, 7)
    net = round(get_dec(row, 8), 2)
    vat = round(get_dec(row, 9), 2)
    gross = round(get_dec(row, 10), 2)

    breakdown = {
        "raw-lines": [],
        "vat": {20: {"vat": vat, "net": net}},
    }
    return {
        "bill_type_code": "N",
        "kwh": Decimal(0),
        "net": net,
        "vat": vat,
        "gross": gross,
        "reads": [],
        "breakdown": breakdown,
        "account": mpan_core,
        "issue_date": issue_date,
        "start_date": start_date,
        "finish_date": finish_date,
        "mpan_core": mpan_core,
        "reference": "_".join(
            (
                start_date.strftime("%Y%m%d"),
                finish_date.strftime("%Y%m%d"),
                issue_date.strftime("%Y%m%d"),
                mpan_core,
            )
        ),
        "elements": [
            {
                "name": "meter",
                "start_date": start_date,
                "finish_date": finish_date,
                "breakdown": {
                    "rate": {meter_rate},
                    "comm": {comm},
                    "settlement-status": {settlement_status},
                    "days": days,
                },
                "net": net,
            }
        ],
    }


class Parser:
    def __init__(self, f):
        self.book = load_workbook(f, data_only=True)
        self.sheet = self.book.worksheets[1]

        self.last_line = None
        self._line_number = None
        self._title_line = None

    @property
    def line_number(self):
        return None if self._line_number is None else self._line_number + 1

    def _set_last_line(self, i, line):
        self._line_numer = i
        self.last_line = line
        if i == 0:
            self._title_line = line
        return line

    def make_raw_bills(self):
        try:
            bills = []
            row = next(self.sheet.iter_rows(min_row=6, max_row=6, max_col=3))
            issue_date = to_utc(get_ct_date(row, 2))
            for row in self.sheet.iter_rows(min_row=12, max_col=11):
                val = get_cell(row, 1).value
                if val is None or val == "":
                    break

                self._set_last_line(row[0].row, val)
                bill = _process_row(issue_date, row)
                bills.append(bill)
        except BadRequest as e:
            raise BadRequest(f"Row number {row[0].row}: {e.description}")

        return bills



import csv
import decimal
from datetime import datetime as Datetime
from decimal import Decimal
from itertools import count

from dateutil.relativedelta import relativedelta

from werkzeug.exceptions import BadRequest

from xlrd import open_workbook, xldate_as_tuple

from chellow.utils import HH, parse_mpan_core, to_ct, to_utc


ELEM_MAP = {
    None: {
        None: None,
        "OOC MOP": {None: "meter-rental"},
    },
    "Charge - Recurring": {None: "duos-fixed"},
    "Meter - Standard": {
        "Energy Bill Relief Scheme": {None: "ebrs"},
        "Energy Bill Relief Scheme Discount": {None: "ebrs"},
        "Energy Bill Discount Scheme": {None: "ebrs"},
    },
    "Meter - UK Electricity - AAHEDC Pass-Thru": {None: "aahedc"},
    "Meter - UK Electricity - BSUoS Pass-Thru": {None: "bsuos"},
    "Meter - UK Electricity - Capacity Market Pass-Thru": {
        None: "capacity",
        "Reverse Capacity Market Estimate": {None: "capacity"},
    },
    "Meter - UK Electricity - CfD FiT Pass-Thru": {None: "cfd-fit"},
    "Meter - UK Electricity - CCL": {
        None: "ccl",
        "CCL": {None: "ccl"},
        "Levy Exempt Energy": {None: "lec"},
    },
    "Meter - UK Electricity - DUoS": {
        None: None,
        "DUoS Unit Rate 3": {
            None: "duos-green",
        },
        "DUoS Unit Charge 3": {
            None: "duos-green",
        },
        "DUoS Unit Charge 2": {
            None: "duos-amber",
        },
        "DUoS Unit Rate 2": {
            None: "duos-amber",
        },
        "DUoS Unit Charge 1": {None: "duos-red"},
        "DUoS Unit Rate 1": {None: "duos-red"},
        "DUoS Standing Charge": {None: "duos-fixed"},
        "DUoS Fixed": {None: "duos-fixed"},
        "DUoS Reactive": {
            None: "duos-reactive",
        },
    },
    "Meter - UK Electricity - FiT Pass-Thru": {
        None: "fit",
    },
    "Pass Thru - UK Electricity Cost Component": {
        None: "meter-rental",
    },
    "Meter - UK Electricity - RO Pass-Thru": {None: "ro"},
    "Meter - UK Electricity - TUoS": {
        None: "triad",
        "TNUoS Fixed": {
            None: "tnuos",
        },
    },
    "Meter - UK Electricity - Standard": {
        None: None,
        "Unit Rate": {
            "Summer Weekday": {
                None: "summer-weekday",
            },
            "Peak": {None: "peak"},
            "Peak Shoulder": {
                None: "peak-shoulder",
            },
            "Summer Night": {
                None: "summer-night",
            },
            "Summer Weekend & Bank Holiday": {
                None: "summer-weekend",
            },
            "Night": {None: "night"},
            "Winter Weekday": {
                None: "winter-weekday",
            },
            "Winter Weekend & Bank Holiday": {
                None: "winter-weekend",
            },
            "Winter Night": {
                None: "winter-night",
            },
            "Day": {None: "day"},
            "Single": {None: "day"},
            "Off Peak / Weekends": {None: "day"},
        },
        "Reverse BSUoS in Unit Rate": {
            None: "bsuos-reverse",
        },
    },
    "Meter - UK Gas - CCL": {None: "ccl"},
}


def _find_name(tree, path):
    if len(path) > 0:
        try:
            return _find_name(tree[path[0]], path[1:])
        except KeyError:
            pass

    return tree[None]


COLUMNS = [
    "Billing Entity",
    "Customer Name",
    "Customer Number",
    "Account Name",
    "Account Number",
    "Billing Address",
    "Bill Number",
    "Bill Date",
    "Due Date",
    "Accepted Date",
    "Bill Period",
    "Agreement Number",
    "Product Bundle",
    "Product Name",
    "Bill Status",
    "Product Item Class",
    "Type",
    "Description",
    "From Date",
    "To Date",
    "Sales Tax Rate",
    "Meter Point",
    "Usage",
    "Usage Unit",
    "Price",
    "Amount",
    "Currency",
    "Indicator",
    "Product Item Name",
    "Rate Name",
]

COLUMN_MAP = dict(zip(COLUMNS, count()))


def get_date(row, name, datemode):
    dt = get_date_naive(row, name, datemode)
    return dt if dt is None else to_utc(to_ct(dt))


def get_date_naive(row, name, datemode):
    val = get_value(row, name)
    if isinstance(val, float):
        return Datetime(*xldate_as_tuple(val, datemode))
    else:
        return None


def get_value(row, name):
    idx = COLUMN_MAP[name]
    try:
        val = row[idx].value
    except IndexError:
        raise BadRequest(
            f"For the name '{name}', the index is {idx} which is beyond the end of "
            f"the row. "
        )
    if isinstance(val, str):
        return val.strip()
    else:
        return val


def get_dec(row, name):
    try:
        return Decimal(str(get_value(row, name)))
    except decimal.InvalidOperation:
        return None


def _get_vat_values(bd, percentage):
    if "vat" in bd:
        vat_breakdown = bd["vat"]
    else:
        vat_breakdown = bd["vat"] = {}

    try:
        vat_values = vat_breakdown[percentage]
    except KeyError:
        vat_values = vat_breakdown[percentage] = {
            "vat": Decimal("0.00"),
            "net": Decimal("0.00"),
        }
    return vat_values


def _parse_row(bills, row, row_index, datemode, title_row):
    val = get_value(row, "Meter Point")
    try:
        mpan_core = parse_mpan_core(str(int(val)))
    except ValueError as e:
        raise BadRequest(
            f"Can't parse the MPAN core in column 'Meter Point' with value '{val}' : "
            f"{e}"
        )

    try:
        mc = bills[mpan_core]
    except KeyError:
        mc = bills[mpan_core] = {}

    bill_period = get_value(row, "Bill Period")
    if "-" in bill_period:
        period_start_naive, period_finish_naive = [
            Datetime.strptime(v, "%Y-%m-%d") for v in bill_period.split(" - ")
        ]
        period_start = to_utc(to_ct(period_start_naive))
        period_finish = to_utc(to_ct(period_finish_naive + relativedelta(days=1) - HH))
    else:
        period_start, period_finish = None, None

    issue_date = get_date(row, "Bill Date", datemode)

    try:
        bill = mc[bill_period]
    except KeyError:
        bill = mc[bill_period] = {
            "bill_type_code": "N",
            "kwh": Decimal(0),
            "vat": Decimal("0.00"),
            "net": Decimal("0.00"),
            "reads": [],
            "breakdown": {"raw_lines": [str(title_row)]},
            "account": mpan_core,
            "issue_date": issue_date,
            "start_date": period_start,
            "finish_date": period_finish,
            "mpan_core": mpan_core,
            "elements": [],
        }

    from_date = get_date(row, "From Date", datemode)
    if from_date is None:
        if period_start is None:
            raise BadRequest("Can't find a bill start date.")
        else:
            from_date = period_start

    to_date_naive = get_date_naive(row, "To Date", datemode)
    if to_date_naive is None:
        if period_finish is None:
            raise BadRequest("Can't find a bill finish date.")
        else:
            to_date = period_finish

    else:
        to_date = to_utc(to_ct(to_date_naive + relativedelta(days=1) - HH))

    bill_number = get_value(row, "Bill Number")
    bd = bill["breakdown"]

    usage = get_dec(row, "Usage")
    usage_units = get_value(row, "Usage Unit")
    price = get_dec(row, "Price")
    amount = get_dec(row, "Amount")
    product_item_name = get_value(row, "Product Item Name")
    rate_name = get_value(row, "Rate Name")
    if product_item_name == "Renewables Obligation (RO)" and usage is not None:
        bill["kwh"] += round(usage, 2)
    description = get_value(row, "Description")
    product_class = get_value(row, "Product Item Class")

    if description in ("Standard VAT@20%", "Reduced VAT@5%"):
        vat_gbp = round(amount, 2)
        bill["vat"] += vat_gbp
        if description.endswith("20%"):
            vat_percentage = Decimal("20")
        else:
            vat_percentage = Decimal("5")
        bd["vat_percentage"] = vat_percentage

        if "vat-rate" in bd:
            vat_rate = bd["vat-rate"]
        else:
            vat_rate = bd["vat-rate"] = set()

        vat_rate.add(vat_percentage / Decimal("100"))

        vat_values = _get_vat_values(bd, vat_percentage)
        vat_values["vat"] += vat_gbp
    else:
        net = round(amount, 2)
        bill["net"] += net

        sales_tax_rate = get_value(row, "Sales Tax Rate")
        if sales_tax_rate == "Commercial UK Energy VAT":
            vat_values = _get_vat_values(bd, 20)
            vat_values["net"] += net

        path = [product_class, description, rate_name]
        elname = _find_name(ELEM_MAP, path)
        ebd = {}
        duos_avail_prefix = "DUoS Availability ("
        duos_excess_avail_prefix = "DUoS Excess Availability ("

        for prefix, name in (
            ("DUoS Availability Adjustment ", "duos-availability"),
            ("DUoS Availability", "duos-availability"),
            ("DUoS Excess Availability", "duos-excess-availability"),
            ("BSUoS Black Start ", "black-start"),
            ("BSUoS Reconciliation - ", "bsuos"),
            ("FiT Rec - ", "fit"),
            ("FiT Reconciliation ", "fit"),
            ("CfD FiT Rec - ", "cfd-fit"),
            ("CfD FiT Reconciliation", "cfd-fit"),
            ("Flex", "reconciliation"),
            ("Legacy TNUoS Reversal ", "triad"),
            ("Hand Held Read -", "meter-rental"),
            ("RO Mutualisation ", "ro"),
            ("OOC MOP - ", "meter-rental"),
            ("KVa Adjustment ", "duos-availability"),
        ):
            if description.startswith(prefix):
                elname = name
                if description.startswith(duos_excess_avail_prefix):
                    ebd["kva"] = int(description[len(duos_excess_avail_prefix) : -5])
                elif description.startswith(duos_avail_prefix):
                    ebd["kva"] = int(description[len(duos_avail_prefix) : -5])
                break

        if elname is None:
            raise BadRequest(
                f"For the path {path} the parser can't work out the element."
            )

        if usage is not None:
            if isinstance(usage_units, str) and usage_units != "":
                units = usage_units.lower()
            else:
                units = "kwh"
            ebd[units] = usage
        if price is not None:
            ebd["rate"] = price
        bill["elements"].append(
            {
                "start_date": from_date,
                "finish_date": to_date,
                "name": elname,
                "net": net,
                "breakdown": ebd,
            }
        )

    reference = str(bill_number) + "_" + str(row_index + 1)

    bill["reference"] = reference
    bill["gross"] = bill["net"] + bill["vat"]


def _make_raw_bills(sheet, datemode):
    bills = {}
    title_row = sheet.row(0)
    for row_index in range(1, sheet.nrows):
        row = sheet.row(row_index)
        if len(row) > 21:
            val = row[21].value
            if val not in (None, ""):
                try:
                    _parse_row(bills, row, row_index, datemode, title_row)
                except BadRequest as e:
                    raise BadRequest(f"On row {row_index + 1}: {e.description}")

    return [b for mc in bills.values() for b in mc.values()]


class Parser:
    def __init__(self, f):
        self.book = open_workbook(file_contents=f.read())
        self.sheet = self.book.sheet_by_index(0)

        self.last_line = None
        lines = (self._set_last_line(i, l) for i, l in enumerate(f))
        self.reader = csv.reader(lines, skipinitialspace=True)
        self._line_number = None
        self._title_line = None

    @property
    def line_number(self):
        return None if self._line_number is None else self._line_number + 1

    def _set_last_line(self, i, line):
        self._line_numer = i
        self.last_line = line
        if i == 0:
            self._title_line = line
        return line

    def make_raw_bills(self):
        return _make_raw_bills(self.sheet, self.book.datemode)



from datetime import datetime as Datetime, timedelta as Timedelta
from decimal import Decimal
from io import BytesIO

from dateutil.relativedelta import relativedelta

from pypdf import PdfReader

from sqlalchemy import null, or_, select

from werkzeug.exceptions import BadRequest

import chellow.e.duos
from chellow.models import Contract, RateScript
from chellow.national_grid import api_get
from chellow.rate_server import download
from chellow.utils import (
    c_months_u,
    ct_datetime,
    hh_after,
    hh_format,
    hh_min,
    to_ct,
    to_utc,
)


def hh(ds, rate_period="monthly", est_kw=None):
    for hh in ds.hh_data:
        if hh["ct-is-month-end"] and ds.dtc_meter_type is not None:
            _process_triad_hh(ds, rate_period, est_kw, hh)


def _find_triad_rate(ds, date, polarity, gsp_group_code):
    trt = ds.non_core_rate("triad_rates", date)["triad_gbp_per_gsp_kw"][polarity][
        gsp_group_code
    ]
    if isinstance(trt, Decimal):
        rt = trt
    elif polarity == "import":
        rt = trt["HHTariff(Floored)_£/kW"]
    elif polarity == "export":
        rt = trt["EET(Floored)_£/kW"]
    return float(rt)


def _process_triad_hh(ds, rate_period, est_kw, hh):
    month_start, month_finish = next(
        c_months_u(start_year=hh["ct-year"], start_month=hh["ct-month"])
    )

    month_start_ct = to_ct(month_start)
    if month_start_ct.month > 3:
        year = month_start_ct.year
    else:
        year = month_start_ct.year - 1
    financial_year_start = to_utc(ct_datetime(year, 4, 1))
    last_financial_year_start = to_utc(ct_datetime(year - 1, 4, 1))
    financial_year_finish = to_utc(ct_datetime(year + 1, 3, 31, 23, 30))

    est_triad_kws = []
    earliest_triad = None
    for dt in ds.non_core_rate("triad_dates", last_financial_year_start)["triad_dates"]:
        triad_hh = None
        earliest_triad = hh_min(earliest_triad, dt)
        try:
            d = next(ds.get_data_sources(dt, dt, financial_year_start))
            chellow.e.duos.duos_vb(d)
            triad_hh = d.hh_data[0]

            fdt = dt
            while fdt < financial_year_start:
                fdt += relativedelta(years=1)

            while fdt.weekday() != dt.weekday():
                fdt += relativedelta(days=1)

            for d in ds.get_data_sources(fdt, fdt, financial_year_start):
                chellow.e.duos.duos_vb(d)
                datum = d.hh_data[0]
                triad_hh["laf"] = datum["laf"]
                triad_hh["gsp-kw"] = datum["laf"] * triad_hh["msp-kw"]
        except StopIteration:
            triad_hh = {
                "hist-start": dt,
                "msp-kw": 0,
                "start-date": dt,
                "status": "before start of MPAN",
                "laf": 1,
                "gsp-kw": 0,
            }
        est_triad_kws.append(triad_hh)

    if ds.site is None:
        era = ds.supply.find_era_at(ds.sess, earliest_triad)
        if (
            era is None or era.get_channel(ds.sess, ds.is_import, "ACTIVE") is None
        ) and est_kw is None:
            est_kw = 0.85 * max(datum["msp-kwh"] for datum in ds.hh_data) * 2
        if est_kw is not None:
            for est_datum in est_triad_kws:
                est_datum["msp-kw"] = est_kw
                est_datum["gsp-kw"] = est_datum["msp-kw"] * est_datum["laf"]

    gsp_kw = 0
    for i, triad_hh in enumerate(est_triad_kws):
        triad_prefix = "triad-estimate-" + str(i + 1)
        hh[triad_prefix + "-date"] = triad_hh["hist-start"]
        hh[triad_prefix + "-msp-kw"] = triad_hh["msp-kw"]
        hh[triad_prefix + "-status"] = triad_hh["status"]
        hh[triad_prefix + "-laf"] = triad_hh["laf"]
        hh[triad_prefix + "-gsp-kw"] = triad_hh["gsp-kw"]
        gsp_kw += triad_hh["gsp-kw"]

    hh["triad-estimate-gsp-kw"] = gsp_kw / 3
    polarity = "import" if ds.llfc.is_import else "export"
    gsp_group_code = ds.gsp_group_code
    rate = _find_triad_rate(ds, month_start, polarity, gsp_group_code)

    hh["triad-estimate-rate"] = rate

    est_triad_gbp = hh["triad-estimate-rate"] * hh["triad-estimate-gsp-kw"]

    if rate_period == "monthly":
        total_intervals = 12

        est_intervals = 1
        hh["triad-estimate-months"] = est_intervals
    else:
        dt = financial_year_start
        total_intervals = 0
        while dt <= financial_year_finish:
            total_intervals += 1
            dt += relativedelta(days=1)

        est_intervals = 0
        for d in ds.get_data_sources(month_start, month_finish):
            for h in d.hh_data:
                if h["ct-decimal-hour"] == 0:
                    est_intervals += 1

        hh["triad-estimate-days"] = est_intervals

    hh["triad-estimate-gbp"] = est_triad_gbp / total_intervals * est_intervals

    if hh["ct-month"] == 3:
        triad_kws = []
        for t_date in ds.non_core_rate("triad_dates", month_start)["triad_dates"]:
            try:
                d = next(ds.get_data_sources(t_date, t_date))
                if (
                    ds.supplier_contract is None
                    or d.supplier_contract == ds.supplier_contract
                ):
                    chellow.e.duos.duos_vb(d)
                    thh = d.hh_data[0]
                else:
                    thh = {
                        "hist-start": t_date,
                        "msp-kw": 0,
                        "start-date": t_date,
                        "status": "before contract",
                        "laf": "before contract",
                        "gsp-kw": 0,
                    }
            except StopIteration:
                thh = {
                    "hist-start": t_date,
                    "msp-kw": 0,
                    "start-date": t_date,
                    "status": "before start of supply",
                    "laf": "before start of supply",
                    "gsp-kw": 0,
                }

            while t_date < financial_year_start:
                t_date += relativedelta(years=1)

            try:
                d = next(ds.get_data_sources(t_date, t_date))
                if (
                    ds.supplier_contract is None
                    or d.supplier_contract == ds.supplier_contract
                ):
                    chellow.e.duos.duos_vb(d)
                    thh["laf"] = d.hh_data[0]["laf"]
                    thh["gsp-kw"] = thh["laf"] * thh["msp-kw"]
            except StopIteration:
                pass

            triad_kws.append(thh)
        gsp_kw = 0

        for i, triad_hh in enumerate(triad_kws):
            pref = "triad-actual-" + str(i + 1)
            hh[pref + "-date"] = triad_hh["start-date"]
            hh[pref + "-msp-kw"] = triad_hh["msp-kw"]
            hh[pref + "-status"] = triad_hh["status"]
            hh[pref + "-laf"] = triad_hh["laf"]
            hh[pref + "-gsp-kw"] = triad_hh["gsp-kw"]
            gsp_kw += triad_hh["gsp-kw"]

        hh["triad-actual-gsp-kw"] = gsp_kw / 3
        polarity = "import" if ds.llfc.is_import else "export"
        gsp_group_code = ds.gsp_group_code
        tot_rate = 0
        for rs in ds.sess.execute(
            select(RateScript)
            .join(Contract, RateScript.contract_id == Contract.id)
            .where(
                Contract.name == "tnuos",
                RateScript.start_date <= financial_year_finish,
                or_(
                    RateScript.finish_date == null(),
                    RateScript.finish_date >= financial_year_start,
                ),
            )
        ).scalars():
            start_month = to_ct(rs.start_date).month
            if start_month < 4:
                start_month += 12

            if rs.finish_date is None:
                finish_month = 3
            else:
                finish_month = to_ct(rs.finish_date).month

            if finish_month < 4:
                finish_month += 12

            rate = _find_triad_rate(ds, rs.start_date, polarity, gsp_group_code)

            tot_rate += (finish_month - start_month + 1) * rate

        rate = tot_rate / 12
        hh["triad-actual-rate"] = rate

        hh["triad-actual-gbp"] = hh["triad-actual-rate"] * hh["triad-actual-gsp-kw"]

        era = ds.supply.find_era_at(ds.sess, month_finish)
        est_intervals = 0

        interval = (
            relativedelta(months=1)
            if rate_period == "monthly"
            else relativedelta(days=1)
        )

        dt = month_finish
        while era is not None and dt > financial_year_start:
            est_intervals += 1
            dt -= interval
            if hh_after(dt, era.finish_date):
                era = ds.supply.find_era_at(ds.sess, dt)

        if rate_period == "monthly":
            hh["triad-all-estimates-months"] = est_intervals
        else:
            hh["triad-all-estimates-days"] = est_intervals
        hh["triad-all-estimates-gbp"] = (
            est_triad_gbp / total_intervals * est_intervals * -1
        )


GSP_LOOKUP = {
    1: "_P",
    2: "_N",
    3: "_F",
    4: "_G",
    5: "_M",
    6: "_D",
    7: "_B",
    8: "_E",
    9: "_A",
    10: "_K",
    11: "_J",
    12: "_C",
    13: "_H",
    14: "_L",
}


def _parse_varying_date(date_str):
    if "-" in date_str:
        pattern = "%Y-%m-%d"
    elif "/" in date_str:
        pattern = "%d/%m/%Y"
    else:
        raise BadRequest(f"The date {date_str} is not recognized.")
    return Datetime.strptime(date_str, pattern)


def national_grid_import(sess, log, set_progress, s):
    log("Starting to check for new TNUoS TRIAD Tariffs")

    contract = Contract.find_non_core_by_name(sess, "triad_rates")
    if contract is None:
        contract = Contract.insert_non_core(
            sess, "triad_rates", "", {}, to_utc(ct_datetime(1996, 4, 1)), None, {}
        )
    for polarity, datafile in (
        ("import", "6abe8416-149c-4c78-b65b-059eea19957a"),
        ("export", "af4501a0-6b96-4088-926c-7c7b0c499b08"),
    ):
        params = {"sql": f"""SELECT * FROM "{datafile}" """}
        res_j = api_get(s, "datastore_search_sql", params=params)
        for record in res_j["result"]["records"]:
            fy_year = int(record["Year_FY"]) - 1
            fy_start = to_utc(ct_datetime(fy_year, 4, 1))
            if fy_start < contract.start_rate_script.start_date:
                continue

            rs = sess.execute(
                select(RateScript).where(
                    RateScript.contract == contract,
                    RateScript.start_date == fy_start,
                )
            ).scalar_one_or_none()
            if rs is None:
                rs = contract.insert_rate_script(sess, fy_start, {})

            rs_script = rs.make_script()
            try:
                tr = rs_script["triad_gbp_per_gsp_kw"]
            except KeyError:
                tr = rs_script["triad_gbp_per_gsp_kw"] = {}

            try:
                rates = tr[polarity]
            except KeyError:
                rates = tr[polarity] = {}

            record_key = GSP_LOOKUP[record["Zone_No"]]
            record_published_date = _parse_varying_date(record["Published_Date"])

            rate = rates.get(record_key)
            if (
                rate is None
                or _parse_varying_date(rate["Published_Date"]) < record_published_date
            ):
                rates[record_key] = record
                rs.update(rs_script)
                sess.commit()

    log("Finished TNUoS TRIAD Tariffs")
    sess.commit()


def _find_triad_dates(file_name, file_like):
    dates = []
    rate_script = {"a_file_name": file_name, "triad_dates": dates}

    reader = PdfReader(file_like)
    in_table = False
    for page in reader.pages:
        for line in page.extract_text().splitlines():
            if in_table:
                date_str, period_str, _ = line.split()
                date = Datetime.strptime(date_str, "%d/%m/%Y")
                delta = Timedelta(minutes=30) * (int(period_str) - 1)
                dates.append(to_utc(to_ct(date + delta)))
                if len(dates) == 3:
                    in_table = False
            else:
                if "Demand (MW)" in line:
                    in_table = True

    return rate_script


def rate_server_import(sess, log, set_progress, s, paths):
    log("Starting to check for new TNUoS triad date PDFs")

    year_entries = {}
    for path, url in paths:
        if len(path) == 4:
            year, utility, rate_type, file_name = path
            if utility == "electricity" and rate_type == "tnuos":
                try:
                    fl_entries = year_entries[year]
                except KeyError:
                    fl_entries = year_entries[year] = {}

                fl_entries[file_name] = url

    for year, year_pdfs in sorted(year_entries.items()):
        year_start = to_utc(ct_datetime(year, 4, 1))
        contract = Contract.get_non_core_by_name(sess, "triad_dates")
        if year_start < contract.start_rate_script.start_date:
            continue
        rs = sess.execute(
            select(RateScript).where(
                RateScript.contract == contract,
                RateScript.start_date == year_start,
            )
        ).scalar_one_or_none()
        if rs is None:
            rs = contract.insert_rate_script(sess, year_start, {})

        if len(year_pdfs) > 0:
            file_name, url = sorted(year_pdfs.items())[-1]

            rs_script = rs.make_script()
            if rs_script.get("a_file_name") != file_name:
                rs.update(_find_triad_dates(file_name, BytesIO(download(s, url))))
                log(f"Updated triad dates rate script for {hh_format(year_start)}")

    log("Finished TNUoS triad dates PDFs")
    sess.commit()



from chellow.models import Contract, MarketRole, Participant
from chellow.rate_server import run_import
from chellow.utils import ct_datetime, to_utc


def test_run_import(mocker, sess):
    vf = to_utc(ct_datetime(1996, 4, 1))
    calb_participant = Participant.insert(sess, "CALB", "AK Industries")
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    calb_participant.insert_party(sess, market_role_Z, "NonCore", vf, None, "")
    Contract.insert_non_core(sess, "configuration", "", {}, vf, None, {})
    sess.commit()
    messages = []

    mocker.patch("chellow.e.bsuos.rate_server_import")
    mocker.patch("chellow.e.dno_rate_parser.rate_server_import")
    mocker.patch("chellow.e.lafs.rate_server_import")
    mocker.patch("chellow.e.mdd_importer.rate_server_import")
    mocker.patch("chellow.e.tlms.rate_server_import")
    mocker.patch("chellow.gas.dn_rate_parser.rate_server_import")

    def mock_api_get(s, url, params=None):
        responses = {
            "https://api.github.com/repos/WessexWater/chellow-rates": {
                "html_url": "https://example.com"
            },
            "https://api.github.com/repos/WessexWater/chellow-rates/branches/main": {
                "commit": {"commit": {"tree": {"url": "https://example.com/tree"}}}
            },
            "https://example.com/tree": {"tree": [], "truncated": False},
        }
        return responses[url]

    mocker.patch("chellow.rate_server.api_get", mock_api_get)

    def log(message):
        messages.append(message)

    def set_progress(progress):
        pass

    run_import(sess, log, set_progress)


def test_run_import_lowercase_readme(mocker, sess):
    vf = to_utc(ct_datetime(1996, 4, 1))
    calb_participant = Participant.insert(sess, "CALB", "AK Industries")
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    calb_participant.insert_party(sess, market_role_Z, "NonCore", vf, None, "")
    Contract.insert_non_core(sess, "configuration", "", {}, vf, None, {})
    sess.commit()
    messages = []

    mocker.patch("chellow.e.bsuos.rate_server_import")
    mocker.patch("chellow.e.dno_rate_parser.rate_server_import")
    mocker.patch("chellow.e.lafs.rate_server_import")
    mocker.patch("chellow.e.mdd_importer.rate_server_import")
    mocker.patch("chellow.e.tlms.rate_server_import")
    mocker.patch("chellow.gas.dn_rate_parser.rate_server_import")

    def mock_api_get(s, url, params=None):
        responses = {
            "https://api.github.com/repos/WessexWater/chellow-rates": {
                "html_url": "https://example.com"
            },
            "https://api.github.com/repos/WessexWater/chellow-rates/branches/main": {
                "commit": {"commit": {"tree": {"url": "https://example.com/tree"}}}
            },
            "https://example.com/tree": {
                "tree": [{"path": "readme.md"}],
                "truncated": False,
            },
        }
        return responses[url]

    mocker.patch("chellow.rate_server.api_get", mock_api_get)

    def log(message):
        messages.append(message)

    def set_progress(progress):
        pass

    run_import(sess, log, set_progress)



from csv import reader
from decimal import Decimal
from io import StringIO


from chellow.models import (
    BillType,
    Comm,
    Contract,
    Cop,
    DtcMeterType,
    EnergisationStatus,
    GspGroup,
    MarketRole,
    MeterPaymentType,
    MeterType,
    Mtc,
    MtcLlfc,
    MtcParticipant,
    Participant,
    Pc,
    Site,
    Source,
    User,
    UserRole,
    VoltageLevel,
    insert_bill_types,
    insert_comms,
    insert_cops,
    insert_dtc_meter_types,
    insert_energisation_statuses,
    insert_generator_types,
    insert_sources,
    insert_voltage_levels,
)
from chellow.reports.report_bills import content
from chellow.utils import ct_datetime, to_utc, utc_datetime


def test_content(mocker, sess):
    vf = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")

    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop Ltd", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion Ltc", vf, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)

    mop_contract = Contract.insert_mop(
        sess, "Fusion Mop Contract", participant, "", {}, vf, None, {}
    )

    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    pc = Pc.insert(sess, "00", "hh", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")

    imp_supplier_contract = Contract.insert_supplier(
        sess, "Fusion Supplier 2000", participant, "", {}, vf, None, {}
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    Contract.insert_dno(sess, dno.dno_code, participant, "", {}, vf, None, {})
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", vf, None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc = Mtc.insert(sess, "845", False, True, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        vf,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess, "510", "PC 5-8 & HH HV", voltage_level, False, True, vf, None
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    insert_sources(sess)
    source_grid = Source.get_by_code(sess, "grid")
    insert_generator_types(sess)
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    supply = site.insert_e_supply(
        sess,
        source_grid,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        dc_contract,
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        dtc_meter_type,
        "22 7867 6232 781",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )

    batch = imp_supplier_contract.insert_batch(sess, "b", "b")
    insert_bill_types(sess)
    bill_type = BillType.get_by_code(sess, "N")
    batch.insert_bill(
        sess,
        "acc",
        "ref",
        to_utc(ct_datetime(2020, 1, 1)),
        to_utc(ct_datetime(2019, 1, 1)),
        to_utc(ct_datetime(2019, 1, 31)),
        Decimal("0"),
        Decimal("0.00"),
        Decimal("0.00"),
        Decimal("0.00"),
        bill_type,
        {"vat": {5: {"net": 3, "vat": 6}}},
        supply,
    )

    editor = UserRole.insert(sess, "editor")
    user = User.insert(sess, "admin@example.com", "xxx", editor, None)
    user_id = user.id

    sess.commit()

    mock_file = StringIO()
    mock_file.close = mocker.Mock()
    mocker.patch("chellow.reports.report_bills.open_file", return_value=mock_file)

    content(user_id, batch.id)
    mock_file.seek(0)
    table = [row for row in reader(mock_file)]

    expected = [
        [
            "chellow_id",
            "supplier_contract",
            "batch_reference",
            "bill_reference",
            "imp_mpan_core",
            "account",
            "issued",
            "from",
            "to",
            "kwh",
            "net",
            "vat",
            "gross",
            "type",
            "vat_1_percent",
            "vat_1_net",
            "vat_1_vat",
            "vat_2_percent",
            "vat_2_net",
            "vat_2_vat",
            "breakdown",
        ],
        [
            "1",
            "Fusion Supplier 2000",
            "b",
            "ref",
            "22 7867 6232 781",
            "acc",
            "2020-01-01 00:00",
            "2019-01-01 00:00",
            "2019-01-31 00:00",
            "0",
            "0.00",
            "0.00",
            "0.00",
            "N",
            "",
            "3",
            "6",
            "",
            "",
            "",
            '{\n  "vat": {\n    5: {\n      "net": 3,'
            '\n      "vat": 6,\n    },\n  },\n}',
        ],
    ]
    print(table)
    assert expected == table



from collections import defaultdict
from decimal import Decimal


from werkzeug.exceptions import BadRequest

from chellow.edi_lib import parse_edi, to_date, to_decimal, to_finish_date
from chellow.utils import HH, parse_mpan_core


read_type_map = {
    "00": "N",
    "09": "N3",
    "04": "C",
    "02": "E",
    "11": "E3",
    "01": "EM",
    "03": "W",
    "06": "X",
    "05": "CP",
    "12": "IF",
}

ELEMENT_MAP = {
    "AAH": {
        "AAHEDC": {
            "139039": ("aahedc", "rate", "kwh"),
        },
    },
    "ADH": {
        "ADHOC": {
            "020330": ("eii", None, None),
            "064305": ("fit", None, None),
            "590346": ("cfd-operational", None, None),
            "493988": ("reconciliation", None, None),
            "637050": ("meter-rental", "rate", "days"),
        },
    },
    "BUS": {
        "BSUOS": {
            "269100": ("bsuos", "rate", "kwh"),
        },
    },
    "CCL": {
        "CCL": {
            "422733": ("ccl", "rate", "kwh"),
        },
    },
    "CFD": {
        "CFD001": {
            "273237": ("cfd-operational", "rate", "kwh"),
            "954379": ("cfd-interim", "rate", "kwh"),
            "538249": ("cm-settlement-levy", "rate", "kwh"),
            "568307": ("capacity", "rate", "kwh"),
        },
    },
    "DCA": {
        "DCDA": {
            "095469": ("meter-rental", "rate", "days"),
        },
    },
    "DUS": {
        "DUS001": {
            "794486": ("duos-availability", "rate", "kva"),
            "644819": ("duos-fixed", "rate", "days"),
            "797790": ("duos-reactive", "rate", "kvarh"),
            "806318": ("duos-green", "rate", "kwh"),
            "716514": ("duos-amber", "rate", "kwh"),
            "769979": ("duos-red", "rate", "kwh"),
            "709522": ("duos-excess-availability", "rate", "kva"),
            "209269": ("tnuos", "rate", "days"),
        },
        "DUS002": {
            "185913": ("duos-yellow", "rate", "kwh"),
            "517270": ("duos-black", "rate", "kwh"),
            "797790": ("duos-reactive", "rate", "kvarh"),
            "806318": ("duos-green", "rate", "kwh"),
            "716514": ("duos-amber", "rate", "kwh"),
            "709522": ("duos-excess-availability", "rate", "kva"),
            "769979": ("duos-red", "rate", "kwh"),
            "644819": ("duos-fixed", "rate", "days"),
            "794486": ("duos-availability", "rate", "kva"),
            "209269": ("tnuos", "rate", "days"),
            "065950": ("eii", "rate", "kwh"),
        },
        "DUSDIS": {
            "122568": ("nrg-gsp-losses", "rate", "kwh"),
        },
        "DUSTRN": {
            "122568": ("nrg-nbp-losses", "rate", "kwh"),
        },
    },
    "ELX": {
        "ELEXON": {
            "489920": ("elexon", "rate", "kwh"),
        },
    },
    "FIT": {
        "FIT_LV": {
            "704107": ("fit", "rate", "kwh"),
        },
    },
    "NRG": {
        "HH0002": {
            "033667": ("management", "rate", "kwh"),
            "091890": ("shape", "rate", "kwh"),
            "122568": ("nrg-msp", "rate", "kwh"),
        },
        "R10001": {
            "700285": ("standing", "rate", "days"),
        },
    },
    "REN": {
        "REN001": {
            "229128": ("ro", "rate", "kwh"),
        },
        "REN002": {
            "019090": ("rego", "rate", "kwh"),
        },
    },
    "TUS": {
        "TNUOS": {
            "012069": ("triad", "rate", "kw"),
        },
    },
}

TPR_LOOKUP = {
    "Day": "00043",
    "Off Peak / Weekends": "00210",
    "Night": "00210",
    "Default Rate": "00043",
    "Single": "00210",
}


def _process_BCD(elements, headers):
    issue_date = to_date(elements["IVDT"][0])
    reference = elements["INVN"][0]
    bill_type_code = elements["BTCD"][0]
    sumo = elements["SUMO"]
    headers["start_date"] = to_date(sumo[0])
    headers["finish_date"] = to_finish_date(sumo[1])

    headers["issue_date"] = issue_date
    headers["bill_type_code"] = bill_type_code
    headers["reference"] = reference
    headers["elements"] = []


def _process_BTL(elements, headers):
    uvlt = elements["UVLT"]
    utva = elements["UTVA"]
    tbtl = elements["TBTL"]

    return {
        "mpan_core": headers["mpan_core"],
        "account": headers["account"],
        "bill_type_code": headers["bill_type_code"],
        "reference": headers["reference"],
        "issue_date": headers["issue_date"],
        "start_date": headers["start_date"],
        "finish_date": headers["finish_date"],
        "kwh": headers["kwh"],
        "net": Decimal("0.00") + to_decimal(uvlt) / Decimal("100"),
        "vat": Decimal("0.00") + to_decimal(utva) / Decimal("100"),
        "gross": Decimal("0.00") + to_decimal(tbtl) / Decimal("100"),
        "breakdown": headers["breakdown"],
        "reads": headers["reads"],
        "elements": headers["elements"],
    }


def _process_CCD1(elements, headers):
    tcod = elements["TCOD"]
    pres_read_date = to_finish_date(elements["PRDT"][0])

    prev_read_date = to_finish_date(elements["PVDT"][0])

    m = elements["MLOC"][0]
    mpan = " ".join((m[13:15], m[15:18], m[18:], m[:2], m[2:6], m[6:10], m[10:13]))

    prrd = elements["PRRD"]
    if len(prrd) < 4:
        return
    pres_read_type = read_type_map[prrd[1]]
    prev_read_type = read_type_map[prrd[3]]

    coefficient = Decimal(elements["ADJF"][1]) / Decimal(100000)
    pres_reading_value = Decimal(prrd[0])
    prev_reading_value = Decimal(prrd[2])
    msn = elements["MTNR"][0]
    tpr_code = elements["TMOD"][0]
    if tpr_code == "kW":
        units = "kW"
        tpr_code = None
    elif tpr_code == "kVA":
        units = "kVA"
        tpr_code = None
    else:
        units = "kWh"
        tpr_code = TPR_LOOKUP[tcod[1]]

    try:
        reads = headers["reads"]
    except KeyError:
        reads = headers["reads"] = []

    reads.append(
        {
            "msn": msn,
            "mpan": mpan,
            "coefficient": coefficient,
            "units": units,
            "tpr_code": tpr_code,
            "prev_date": prev_read_date,
            "prev_value": prev_reading_value,
            "prev_type_code": prev_read_type,
            "pres_date": pres_read_date,
            "pres_value": pres_reading_value,
            "pres_type_code": pres_read_type,
        }
    )


def _process_CCD2(elements, headers):
    breakdown = defaultdict(int)

    element_code = elements["TMOD"][0]
    headers["element_code"] = element_code
    try:
        eln_gbp, eln_rate, eln_cons = ELEMENT_MAP[element_code]
    except KeyError:
        raise BadRequest(
            f"Can't find the element code {element_code} in the ELEMENT_MAP."
        )

    cons = elements["CONS"]
    kwh = Decimal("0")
    if eln_cons is not None and len(cons[0]) > 0:
        el_cons = to_decimal(cons) / Decimal("1000")
        if eln_gbp == "duos-availability-gbp":
            breakdown[eln_cons] = [el_cons]
        else:
            breakdown[eln_cons] = kwh = el_cons

    if eln_rate is not None:
        rate = to_decimal(elements["BPRI"]) / Decimal("100000")
        breakdown[eln_rate] = [rate]

    start_date = to_date(elements["CSDT"][0])
    headers["bill_start_date"] = start_date

    finish_date = to_date(elements["CEDT"][0]) - HH
    headers["bill_finish_date"] = finish_date

    if "CTOT" in elements:
        net = Decimal("0.00") + to_decimal(elements["CTOT"]) / Decimal("100")
    else:
        net = Decimal("0.00")

    breakdown[eln_gbp] = net
    breakdown["raw-lines"] = [headers["line"]]

    try:
        reads = headers["reads"]
        headers["reads"] = []
    except KeyError:
        reads = []

    bill = {
        "bill_type_code": headers["bill_type_code"],
        "reference": headers["reference"] + "_" + eln_gbp[:-4],
        "issue_date": headers["issue_date"],
        "start_date": start_date,
        "finish_date": finish_date,
        "kwh": kwh if eln_gbp == "ro-gbp" else Decimal("0"),
        "net": net,
        "vat": Decimal("0.00"),
        "gross": net,
        "breakdown": breakdown,
        "reads": reads,
    }
    headers["bills"].append(bill)


def _process_CCD3(elements, headers):
    elem_bd = {}

    supplier_code = elements["CCDE"][2]
    tariff_code = elements["TCOD"][0]
    mod_code = elements["TMOD"][0]
    try:
        eln_name, eln_rate, eln_cons = ELEMENT_MAP[supplier_code][tariff_code][mod_code]
    except KeyError:
        raise BadRequest(
            f"Can't find the element key {supplier_code} -> {tariff_code} -> "
            f"{mod_code} in the ELEMENT_MAP."
        )

    cons = elements["CONS"]
    if len(cons[0]) > 0:
        el_cons = to_decimal(cons) / Decimal("1000")
        elem_bd[eln_cons] = el_cons
        if eln_name == "nrg-msp":
            headers["kwh"] += el_cons

    bpri = elements["BPRI"]
    if len(bpri[0]) > 0:
        rate = to_decimal(bpri) / Decimal("100000")
        elem_bd[eln_rate] = {rate}

    net = Decimal("0.00")
    if "CTOT" in elements:
        net += to_decimal(elements["CTOT"]) / Decimal("100")

    csdt = elements["CSDT"]
    start_date = to_date(csdt[0])
    cedt = elements["CEDT"]
    finish_date = to_finish_date(cedt[0])

    headers["elements"].append(
        {
            "name": eln_name,
            "start_date": start_date,
            "finish_date": finish_date,
            "net": net,
            "breakdown": elem_bd,
        }
    )


def _process_CCD4(elements, headers):
    breakdown = headers["breakdown"]

    element_code = elements["TMOD"][0]
    headers["element_code"] = element_code
    try:
        eln_gbp, eln_rate, eln_cons = ELEMENT_MAP[element_code]
    except KeyError:
        raise BadRequest(f"Can't find the element code {element_code} in the TMOD_MAP.")

    cons = elements["CONS"]
    if eln_cons is not None and len(cons[0]) > 0:
        el_cons = to_decimal(cons, "1000")
        breakdown[eln_cons] = kwh = el_cons

    if eln_rate is not None:
        rate = to_decimal(elements["BPRI"], "100000")
        breakdown[eln_rate] = [rate]

    start_date = to_date(elements["CSDT"][0])
    headers["bill_start_date"] = start_date

    finish_date = to_date(elements["CEDT"][0]) - HH
    headers["bill_finish_date"] = finish_date

    if "CTOT" in elements:
        net = Decimal("0.00") + to_decimal(elements["CTOT"], "100")
    else:
        net = Decimal("0.00")

    breakdown[eln_gbp] = net
    breakdown["raw-lines"] = [headers["line"]]

    try:
        reads = headers["reads"]
        del headers["reads"][:]
    except KeyError:
        reads = []

    bill = {
        "kwh": kwh if eln_gbp == "ro-gbp" else Decimal("0.00"),
        "reference": headers["reference"] + "_" + eln_gbp[:-4],
        "issue_date": headers["issue_date"],
        "start_date": start_date,
        "finish_date": finish_date,
        "net": net,
        "vat": Decimal("0.00"),
        "gross": net,
        "breakdown": breakdown,
        "reads": reads,
        "bill_type_code": headers["bill_type_code"],
    }
    headers["bills"].append(bill)


def _process_CDT(elements, headers):
    customer_id = elements["CIDN"][0]
    headers["customer_number"] = customer_id


def _process_CLO(elements, headers):
    cloc = elements["CLOC"]
    headers["account"] = cloc[1]


def _process_END(elements, headers):
    pass


def _process_MAN(elements, headers):
    madn = elements["MADN"]

    headers["mpan_core"] = parse_mpan_core("".join(madn[0:3]))


def _process_MHD(elements, headers):
    message_type = elements["TYPE"][0]
    if message_type == "UTLBIL":
        last_line = headers["lines"][-1]
        keep_keys = {"customer_number"}
        keep = {k: headers[k] for k in keep_keys}
        headers.clear()
        headers.update(keep)
        headers["lines"] = [last_line]
        headers["breakdown"] = defaultdict(
            int, {"vat": {}, "raw-lines": headers["lines"]}
        )
        headers["kwh"] = Decimal(0)
        headers["reads"] = []


def _process_MTR(elements, headers):
    headers["lines"] = []


def _process_VAT(elements, headers):
    breakdown = headers["breakdown"]
    vat = Decimal("0.00") + to_decimal(elements["UVTT"]) / Decimal("100")
    vat_percentage = to_decimal(elements["VATP"]) / Decimal("1000")

    if "vat_rate" in breakdown:
        vat_rate = breakdown["vat-rate"]
    else:
        vat_rate = breakdown["vat-rate"] = set()
    vat_rate.add(vat_percentage / Decimal("100"))
    vat_net = Decimal("0.00") + to_decimal(elements["UVLA"]) / Decimal("100")

    breakdown["vat"][vat_percentage] = {"vat": vat, "net": vat_net}


def _process_NOOP(elements, headers):
    pass


CODE_FUNCS = {
    "BCD": _process_BCD,
    "BTL": _process_BTL,
    "CCD1": _process_CCD1,
    "CCD2": _process_CCD2,
    "CCD3": _process_CCD3,
    "CCD4": _process_CCD4,
    "CDT": _process_CDT,
    "CLO": _process_CLO,
    "DNA": _process_NOOP,
    "END": _process_END,
    "FIL": _process_NOOP,
    "MAN": _process_MAN,
    "MHD": _process_MHD,
    "MTR": _process_MTR,
    "SDT": _process_NOOP,
    "STX": _process_NOOP,
    "TYP": _process_NOOP,
    "TTL": _process_NOOP,
    "VAT": _process_VAT,
    "VTS": _process_NOOP,
}


class Parser:
    def __init__(self, f):
        self.edi_str = str(f.read(), "utf-8", errors="ignore")
        self.line_number = None

    def make_raw_bills(self):
        bills = []
        headers = {"lines": []}
        for self.line_number, line, seg_name, elements in parse_edi(self.edi_str):
            headers["lines"].append(line)
            try:
                func = CODE_FUNCS[seg_name]
            except KeyError:
                raise BadRequest(f"Code {seg_name} not recognized.")

            try:
                bill = func(elements, headers)
            except BadRequest as e:
                raise BadRequest(
                    f"{e.description} on line {self.line_number} line {line} "
                    f"seg_name {seg_name} elements {elements}"
                )
            except BaseException as e:
                raise BadRequest(
                    f"{e} on line {self.line_number} line {line} "
                    f"seg_name {seg_name} elements {elements}"
                ) from e

            if bill is not None:
                bills.append(bill)

        return bills



from io import BytesIO

from sqlalchemy import select

from utils import match, match_repeat

from chellow.models import (
    Comm,
    Contract,
    Cop,
    DtcMeterType,
    EnergisationStatus,
    GContract,
    GDn,
    GReadingFrequency,
    GUnit,
    GspGroup,
    MarketRole,
    MeterPaymentType,
    MeterType,
    Mtc,
    MtcLlfc,
    MtcLlfcSsc,
    MtcLlfcSscPc,
    MtcParticipant,
    MtcSsc,
    Participant,
    Pc,
    Report,
    ReportRun,
    Scenario,
    Site,
    Snag,
    Source,
    Ssc,
    VoltageLevel,
    insert_comms,
    insert_cops,
    insert_dtc_meter_types,
    insert_energisation_statuses,
    insert_g_reading_frequencies,
    insert_g_units,
    insert_sources,
    insert_voltage_levels,
)
from chellow.utils import ct_datetime, to_utc, utc_datetime


def test_home(client, sess, mocker):
    vf = to_utc(ct_datetime(2000, 1, 1))

    participant = Participant.insert(sess, "CALB", "AK Industries")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    sess.commit()

    patched_hh_importer = mocker.patch("chellow.chellow.e.hh_importer")

    class AnObj:
        pass

    tsk = AnObj()
    tsk.is_error = True
    tsk.contract_id = dc_contract.id
    patched_hh_importer.tasks = {"a": tsk}

    response = client.get("/")

    match(response, 200, f"/e/dc_contracts/{dc_contract.id}/auto_importer")


def test_input_date(client, sess, mocker):
    query_string = {
        "prefix": "timestamp",
        "resolution": "day",
        "timestamp_year": "2022",
        "timestamp_month": "11",
        "timestamp_day": "31",
    }
    response = client.get("/input_date", query_string=query_string)

    match(response, 200)


def test_local_report_home(client, sess):
    script = """from flask import render_template_string

def do_get():
    return render_template_string(template)
"""

    template = """{% extends "base.html"%}

{% block content %}
  <p>Marcus Aurelius</p>
{% endblock %}
"""

    report = Report.insert(sess, "Emperor", script, template)
    config = sess.execute(
        select(Contract).where(Contract.name == "configuration")
    ).scalar_one()
    props = config.make_properties()
    props["local_reports_id"] = report.id
    config.properties = props
    sess.commit()

    response = client.get("/local_reports_home")

    match(response, 200)


def test_local_report_post_delete(client, sess):
    report = Report.insert(sess, "Gray", "", "")
    sess.commit()

    data = {
        "delete": "Delete",
    }
    response = client.post(f"/local_reports/{report.id}", data=data)

    match(response, 303)


def test_local_reports_post(client, sess):
    data = {
        "name": "Minority Report",
        "add": "Add",
    }
    response = client.post("/local_reports", data=data)

    match(response, 303)

    report = sess.execute(select(Report)).scalar_one()
    assert report.template == ""


def test_scenario_get(sess, client):
    props = {
        "scenario_start_year": 2010,
        "scenario_start_month": 5,
    }
    scenario = Scenario.insert(sess, "scenario 1", props)
    sess.commit()

    response = client.get(f"/scenarios/{scenario.id}")

    match(response, 200)


def test_scenario_edit_post(sess, client):
    props = {
        "scenario_start_year": 2010,
        "scenario_start_month": 5,
        "scenario_duration": 2,
    }
    scenario = Scenario.insert(sess, "scenario 1", props)
    sess.commit()

    data = {
        "name": "scenario_bau",
        "properties": """
{
  "local_rates": [],
  "scenario_start_year": 2015,
  "scenario_start_month": 6,
  "scenario_duration": 1
}""",
    }

    response = client.post(f"/scenarios/{scenario.id}/edit", data=data)

    match(response, 303, r"/scenarios/1")


def test_site_get(client, sess):
    valid_from = to_utc(ct_datetime(2000, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")

    market_role_Z = MarketRole.get_by_code(sess, "Z")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(
        sess, market_role_Z, "None core", utc_datetime(2000, 1, 1), None, None
    )
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess,
        "bank_holidays",
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        bank_holiday_rate_script,
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(
        sess, market_role_M, "Fusion Mop Ltd", utc_datetime(2000, 1, 1), None, None
    )
    participant.insert_party(
        sess, market_role_X, "Fusion Ltc", utc_datetime(2000, 1, 1), None, None
    )
    participant.insert_party(
        sess, market_role_C, "Fusion DC", utc_datetime(2000, 1, 1), None, None
    )
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    pc = Pc.insert(sess, "00", "hh", utc_datetime(2000, 1, 1), None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    imp_supplier_contract = Contract.insert_supplier(
        sess,
        "Fusion Supplier 2000",
        participant,
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        {},
    )
    dno = participant.insert_party(
        sess, market_role_R, "WPD", utc_datetime(2000, 1, 1), None, "22"
    )
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", utc_datetime(2000, 1, 1), None)
    meter_payment_type = MeterPaymentType.insert(
        sess, "CR", "Credit", utc_datetime(1996, 1, 1), None
    )
    mtc = Mtc.insert(
        sess,
        "845",
        False,
        True,
        utc_datetime(1996, 1, 1),
        None,
    )
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        utc_datetime(1996, 1, 1),
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess,
        "510",
        "PC 5-8 & HH HV",
        voltage_level,
        False,
        True,
        utc_datetime(1996, 1, 1),
        None,
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, valid_from, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        utc_datetime(2020, 1, 1),
        gsp_group,
        mop_contract,
        dc_contract,
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        dtc_meter_type,
        "22 0470 7514 535",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    sess.commit()

    response = client.get(f"/sites/{site.id}")

    patterns = []
    match(response, 200, *patterns)


def test_site_get_dumb(client, sess):
    valid_from = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")

    market_role_Z = MarketRole.get_by_code(sess, "Z")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(
        sess, market_role_Z, "None core", utc_datetime(2000, 1, 1), None, None
    )
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess,
        "bank_holidays",
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        bank_holiday_rate_script,
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(
        sess, market_role_M, "Fusion Mop Ltd", utc_datetime(2000, 1, 1), None, None
    )
    participant.insert_party(
        sess, market_role_X, "Fusion Ltc", utc_datetime(2000, 1, 1), None, None
    )
    participant.insert_party(
        sess, market_role_C, "Fusion DC", utc_datetime(2000, 1, 1), None, None
    )
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    pc = Pc.insert(sess, "02", "nhh", utc_datetime(2000, 1, 1), None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    ssc = Ssc.insert(sess, "0001", "All", True, utc_datetime(1996, 1), None)
    imp_supplier_contract = Contract.insert_supplier(
        sess,
        "Fusion Supplier 2000",
        participant,
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        {},
    )
    dno = participant.insert_party(
        sess, market_role_R, "WPD", utc_datetime(2000, 1, 1), None, "22"
    )
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", utc_datetime(2000, 1, 1), None)
    meter_payment_type = MeterPaymentType.insert(
        sess, "CR", "Credit", utc_datetime(1996, 1, 1), None
    )
    mtc = Mtc.insert(
        sess,
        "845",
        False,
        True,
        valid_from,
        None,
    )
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        utc_datetime(1996, 1, 1),
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess,
        "510",
        "PC 5-8 & HH HV",
        voltage_level,
        False,
        True,
        utc_datetime(1996, 1, 1),
        None,
    )
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    MtcLlfc.insert(sess, mtc_participant, llfc, valid_from, None)
    mtc_ssc = MtcSsc.insert(sess, mtc_participant, ssc, valid_from, None)
    mtc_llfc_ssc = MtcLlfcSsc.insert(sess, mtc_ssc, llfc, valid_from, None)
    MtcLlfcSscPc.insert(sess, mtc_llfc_ssc, pc, valid_from, None)
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        utc_datetime(2020, 1, 1),
        gsp_group,
        mop_contract,
        dc_contract,
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        ssc.code,
        energisation_status,
        dtc_meter_type,
        "22 0470 7514 535",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    sess.commit()

    response = client.get(f"/sites/{site.id}")

    patterns = []
    match(response, 200, *patterns)


def test_supplies_get_e(sess, client):
    valid_from = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")

    market_role_Z = MarketRole.get_by_code(sess, "Z")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", valid_from, None, None)
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess,
        "bank_holidays",
        "",
        {},
        valid_from,
        None,
        bank_holiday_rate_script,
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(
        sess, market_role_M, "Fusion Mop Ltd", valid_from, None, None
    )
    participant.insert_party(
        sess, market_role_X, "Fusion Ltc", utc_datetime(2000, 1, 1), None, None
    )
    participant.insert_party(
        sess, market_role_C, "Fusion DC", utc_datetime(2000, 1, 1), None, None
    )
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    pc = Pc.insert(sess, "02", "nhh", utc_datetime(2000, 1, 1), None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    ssc = Ssc.insert(sess, "0001", "All", True, utc_datetime(1996, 1), None)
    imp_supplier_contract = Contract.insert_supplier(
        sess,
        "Fusion Supplier 2000",
        participant,
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        {},
    )
    dno = participant.insert_party(
        sess, market_role_R, "WPD", utc_datetime(2000, 1, 1), None, "22"
    )
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", utc_datetime(2000, 1, 1), None)
    meter_payment_type = MeterPaymentType.insert(
        sess, "CR", "Credit", utc_datetime(1996, 1, 1), None
    )
    mtc = Mtc.insert(
        sess,
        "845",
        False,
        True,
        valid_from,
        None,
    )
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        utc_datetime(1996, 1, 1),
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess,
        "510",
        "PC 5-8 & HH HV",
        voltage_level,
        False,
        True,
        utc_datetime(1996, 1, 1),
        None,
    )
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    MtcLlfc.insert(sess, mtc_participant, llfc, valid_from, None)
    mtc_ssc = MtcSsc.insert(sess, mtc_participant, ssc, valid_from, None)
    mtc_llfc_ssc = MtcLlfcSsc.insert(sess, mtc_ssc, llfc, valid_from, None)
    MtcLlfcSscPc.insert(sess, mtc_llfc_ssc, pc, valid_from, None)
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        utc_datetime(2020, 1, 1),
        gsp_group,
        mop_contract,
        dc_contract,
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        ssc.code,
        energisation_status,
        dtc_meter_type,
        "22 0470 7514 535",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    site.insert_e_supply(
        sess,
        source,
        None,
        "Alice",
        utc_datetime(2000, 1, 1),
        utc_datetime(2020, 1, 1),
        gsp_group,
        mop_contract,
        dc_contract,
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        ssc.code,
        energisation_status,
        dtc_meter_type,
        "22 0471 7514 532",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    sess.commit()
    query_string = {
        "search_pattern": "",
    }
    response = client.get("/supplies", query_string=query_string)

    match(response, 200)


def test_supplies_get_g(sess, client):
    vf = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "22488", "Water Works")

    g_dn = GDn.insert(sess, "EE", "East of England")

    g_ldz = g_dn.insert_g_ldz(sess, "EA")

    g_exit_zone = g_ldz.insert_g_exit_zone(sess, "EA1")

    insert_g_units(sess)

    g_unit_M3 = GUnit.get_by_code(sess, "M3")

    g_contract = GContract.insert(sess, False, "Fusion 2020", "", {}, vf, None, {})

    insert_g_reading_frequencies(sess)

    g_reading_frequency_M = GReadingFrequency.get_by_code(sess, "M")

    mprn = "87614362"
    g_supply = site.insert_g_supply(
        sess,
        mprn,
        "main",
        g_exit_zone,
        utc_datetime(2018, 1, 1),
        None,
        "hgeu8rhg",
        1,
        g_unit_M3,
        g_contract,
        "d7gthekrg",
        g_reading_frequency_M,
        1,
        1,
    )
    sess.commit()
    query_string = {
        "search_pattern": mprn,
    }
    response = client.get("/supplies", query_string=query_string)

    match(response, 307, rf"/g/supplies/{g_supply.id}")


def test_system(client):
    response = client.get("/system")

    match(response, 200)


def test_general_import_post_full(sess, client):
    """General import of channel snag unignore and check the import that's
    been created.
    """
    valid_from = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")

    market_role_Z = MarketRole.get_by_code(sess, "Z")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(
        sess, market_role_Z, "None core", utc_datetime(2000, 1, 1), None, None
    )
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess,
        "bank_holidays",
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        bank_holiday_rate_script,
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(
        sess, market_role_M, "Fusion Mop Ltd", utc_datetime(2000, 1, 1), None, None
    )
    participant.insert_party(
        sess, market_role_X, "Fusion Ltc", utc_datetime(2000, 1, 1), None, None
    )
    participant.insert_party(
        sess, market_role_C, "Fusion DC", utc_datetime(2000, 1, 1), None, None
    )
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    pc = Pc.insert(sess, "00", "hh", utc_datetime(2000, 1, 1), None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    imp_supplier_contract = Contract.insert_supplier(
        sess,
        "Fusion Supplier 2000",
        participant,
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        {},
    )
    dno = participant.insert_party(
        sess, market_role_R, "WPD", utc_datetime(2000, 1, 1), None, "22"
    )
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", utc_datetime(2000, 1, 1), None)
    meter_payment_type = MeterPaymentType.insert(
        sess, "CR", "Credit", utc_datetime(1996, 1, 1), None
    )
    mtc = Mtc.insert(
        sess,
        "845",
        False,
        True,
        utc_datetime(1996, 1, 1),
        None,
    )
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        utc_datetime(1996, 1, 1),
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess,
        "510",
        "PC 5-8 & HH HV",
        voltage_level,
        False,
        True,
        utc_datetime(1996, 1, 1),
        None,
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, valid_from, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    supply = site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        utc_datetime(2020, 1, 1),
        gsp_group,
        mop_contract,
        dc_contract,
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        dtc_meter_type,
        "22 0470 7514 535",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    era = supply.eras[0]
    channel = era.insert_channel(sess, False, "ACTIVE")
    channel.add_snag(
        sess,
        "Missing",
        utc_datetime(2003, 8, 2, 23, 30),
        utc_datetime(2004, 7, 6, 22, 30),
    )
    snag = sess.query(Snag).order_by(Snag.start_date).first()
    snag.set_is_ignored(True)
    sess.commit()

    file_items = [
        "insert",
        "channel_snag_unignore",
        "22 0470 7514 535",
        "FALSE",
        "Active",
        "Missing",
        "2003-08-03 00:00",
        "2004-07-06 23:30",
    ]
    file_name = "gi_channel_snag_ignore.csv"
    file_bytes = ",".join(file_items).encode("utf8")
    f = BytesIO(file_bytes)

    data = {"import_file": (f, file_name)}

    response = client.post("/general_imports", data=data)

    match(response, 303, "/general_imports/0")

    match_repeat(
        client, "/general_imports/0", r"The file has been imported successfully"
    )
    sess.rollback()

    assert not snag.is_ignored


def test_non_core_auto_importer_get(sess, client):
    market_role_Z = MarketRole.get_by_code(sess, "Z")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(
        sess, market_role_Z, "None core", utc_datetime(2000, 1, 1), None, None
    )
    contract = Contract.insert_non_core(
        sess, "bank_holidays", "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    sess.commit()

    response = client.get(f"/non_core_contracts/{contract.id}/auto_importer")

    match(response, 200)


def test_non_core_auto_importer_get_bmarketidx(sess, client):
    vf = to_utc(ct_datetime(2000, 1, 1))
    market_role_Z = MarketRole.get_by_code(sess, "Z")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    contract = Contract.insert_non_core(sess, "bmarketidx", "", {}, vf, None, {})
    sess.commit()

    response = client.get(f"/non_core_contracts/{contract.id}/auto_importer")

    match(response, 200)


def test_non_core_auto_importer_post(mocker, sess, client):
    market_role_Z = MarketRole.get_by_code(sess, "Z")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(
        sess, market_role_Z, "None core", utc_datetime(2000, 1, 1), None, None
    )
    contract = Contract.insert_non_core(
        sess, "bank_holidays", "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    sess.commit()

    mocker.patch("chellow.views.import_module")

    response = client.post(f"/non_core_contracts/{contract.id}/auto_importer")

    match(response, 303)


def test_non_core_contract_edit_post(sess, client):
    market_role_Z = MarketRole.get_by_code(sess, "Z")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(
        sess, market_role_Z, "None core", utc_datetime(2000, 1, 1), None, None
    )
    contract = Contract.insert_non_core(
        sess, "g_cv", "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    sess.commit()

    data = {
        "name": "g_cv",
        "properties": """{
"enabled": true,
"url": "http://localhost:8080/nationalgrid/cv.csv"}
""",
    }

    response = client.post(f"/non_core_contracts/{contract.id}/edit", data=data)

    match(response, 303)


def test_non_core_rate_script_edit_get(sess, client):
    market_role_Z = MarketRole.get_by_code(sess, "Z")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(
        sess, market_role_Z, "None core", utc_datetime(2000, 1, 1), None, None
    )
    contract = Contract.insert_non_core(
        sess, "rcrc", "import nonexistent", {}, utc_datetime(2000, 1, 1), None, {}
    )
    rs = contract.start_rate_script
    sess.commit()

    response = client.get(f"/non_core_rate_scripts/{rs.id}/edit")

    match(response, 200)


def test_non_core_rate_script_edit_get_bst(sess, client):
    market_role_Z = MarketRole.get_by_code(sess, "Z")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(
        sess, market_role_Z, "None core", utc_datetime(2000, 1, 1), None, None
    )
    contract = Contract.insert_non_core(
        sess, "rcrc", "", {}, to_utc(ct_datetime(2000, 5, 31, 23, 30)), None, {}
    )
    rs = contract.start_rate_script
    sess.commit()

    response = client.get(f"/non_core_rate_scripts/{rs.id}/edit")

    match(response, 200, r'<input type="hidden" name="start_hour" value="23">')


def test_site_edit_get(sess, client):
    vf = to_utc(ct_datetime(2000, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")

    GContract.insert_industry(sess, "cv", "", {}, vf, None, {})
    sess.commit()

    response = client.get(f"/sites/{site.id}/edit")
    patterns = [
        r'<select name="g_contract_id">\s*</select>',
    ]
    match(response, 200, *patterns)


def test_rate_server_get(sess, client):
    response = client.get("/rate_server")
    match(response, 200)


def test_report_run_get_bill_check(sess, client):
    report_run = ReportRun.insert(sess, "bill_check", None, "_b_88", {})
    sess.commit()
    query_string = {"element": "net"}
    response = client.get(f"/report_runs/{report_run.id}", query_string=query_string)
    match(response, 200)


def test_report_run_spreadsheet_get(sess, client):
    report_run = ReportRun.insert(sess, "bill_check", None, "_b_88", {})
    report_run.insert_row(sess, "", ["clump"], {}, {})
    sess.commit()

    response = client.get(f"/report_runs/{report_run.id}/spreadsheet")
    match(response, 200)


def test_report_run_row_get_bill_check(sess, client):
    report_run = ReportRun.insert(sess, "bill_check", None, "_b_88", {})
    report_run_row = report_run.insert_row(
        sess,
        "",
        ["clump"],
        {},
        {},
        data={
            "actual_net_gbp": 10,
            "virtual_net_gbp": 20.32,
            "difference_net_gbp": 23.6,
            "elements": {},
        },
    )
    sess.commit()

    response = client.get(f"/report_run_rows/{report_run_row.id}")
    match(response, 200)



import csv
from collections import defaultdict
from datetime import datetime as Datetime
from decimal import Decimal, InvalidOperation
from enum import Enum, auto
from io import BytesIO

from dateutil.relativedelta import relativedelta

from openpyxl import load_workbook

from werkzeug.exceptions import BadRequest

from chellow.utils import hh_max, hh_min, to_ct, to_utc


class Title(Enum):
    CUSTOMER = auto()
    PRODUCT = auto()
    BROKER_NAME = auto()
    ACCOUNT = auto()
    MPRN = auto()
    BILL = auto()
    SUPPLY_ADDRESS = auto()
    BILL_DATE = auto()
    BILLING_PERIOD = auto()
    CHARGE_TYPE = auto()
    CHARGE_PERIOD_FROM = auto()
    CHARGE_PERIOD_END = auto()
    QUANTITY = auto()
    QUNIT = auto()
    CHARGE = auto()
    CUNIT = auto()
    TOTAL = auto()


COLUMNS = {
    Title.CUSTOMER: ["Customer"],
    Title.ACCOUNT: ["Account"],
    Title.MPRN: ["MPRN"],
    Title.BILL: ["Bill"],
    Title.BILL_DATE: ["Bill Date", "BillDate"],
    Title.BILLING_PERIOD: ["Billing Period", "BillingPeriod"],
    Title.CHARGE_TYPE: ["Charge Type", "ChargeType"],
    Title.CHARGE_PERIOD_FROM: ["Charge Period From", "ChargePeriodFrom"],
    Title.CHARGE_PERIOD_END: ["Charge Period End", "ChargePeriodEnd"],
    Title.QUANTITY: ["Quantity"],
    Title.QUNIT: ["QUnit"],
    Title.CHARGE: ["Charge"],
    Title.CUNIT: ["CUnit"],
    Title.TOTAL: ["Total"],
}


def make_column_map(title_row):
    titles = [cell.value for cell in title_row]
    column_map = {}
    for title, title_names in COLUMNS.items():
        idx = None
        for title_name in title_names:
            try:
                idx = titles.index(title_name)
            except ValueError:
                pass
        if idx is None:
            raise BadRequest(f"For the title {title} a column can't be found")

        column_map[title] = idx + 1
    return column_map


def get_date_naive(sheet, row, col):
    value = get_value(sheet, row, col)
    if value in ("", None):
        return None
    elif not isinstance(value, Datetime):
        raise BadRequest(
            f"Problem reading {value} as a timestamp at row {row} col {col}."
        )
    return value


def get_date(sheet, row, col_name):
    dt = get_date_naive(sheet, row, col_name)
    return None if dt is None else to_utc(to_ct(dt))


def get_value(sheet, row, col):
    try:
        return sheet.cell(row=row, column=col).value
    except IndexError:
        raise BadRequest(
            f"Can't find the cell at row {row} and col {col} on sheet {sheet}."
        )


def get_str(sheet, row, col):
    value = get_value(sheet, row, col)
    if value is None:
        return None

    return value.strip()


def get_dec(sheet, row, col):
    value = get_value(sheet, row, col)
    if value in ("", None):
        return None

    try:
        return Decimal(str(value))
    except InvalidOperation as e:
        raise BadRequest(
            f"Problem parsing the number '{value}' at row {row} col {col}. {e}"
        )


def get_int(sheet, row, col):
    value = get_value(sheet, row, col)
    try:
        return int(value)
    except ValueError as e:
        raise BadRequest(
            f"Problem parsing the integer '{value}' at row {row} col {col}. {e}"
        )


ELEMENT_LOOKUP = {
    "CCL": "ccl",
    "Fixed Management Fee": "admin_fixed",
    "Gas": "admin_variable",
    "Management Fee": "admin_variable",
    "Metering and Data": "metering",
    "LDZ Customer Capacity": "dn_customer_capacity",
    "LDZ Customer Fixed": "dn_customer_fixed",
    "LDZ System Capacity": "dn_system_capacity",
    "LDZ System Commodity": "dn_system_commodity",
    "NTS Exit Capacity (ECN)": "dn_ecn",
    "NTS SO Exit": "so_exit_commodity",
    "NTS TO Exit": "to_exit_commodity",
    "Unidentified Gas": "ug",
    "WAP": "commodity",
}

QUNIT_LOOKUP = {
    "kWh": "kwh",
    "days": "days",
}


def _parse_row(bills, sheet, row, title_row):
    column_map = make_column_map(title_row)
    mprn_raw = get_value(sheet, row, column_map[Title.MPRN])
    mprn = str(mprn_raw)
    reference = str(get_value(sheet, row, column_map[Title.BILL]))
    account = get_value(sheet, row, column_map[Title.ACCOUNT])
    issue_date = get_date(sheet, row, column_map[Title.BILL_DATE])
    charge_period_from = get_date(sheet, row, column_map[Title.CHARGE_PERIOD_FROM])
    charge_period_end_naive = get_date_naive(
        sheet, row, column_map[Title.CHARGE_PERIOD_END]
    )
    if charge_period_end_naive is None:
        charge_period_end = None
    else:
        charge_period_end = to_utc(
            to_ct(charge_period_end_naive) + relativedelta(hours=23, minutes=30)
        )

    try:
        mprn_values = bills[mprn]
    except KeyError:
        mprn_values = bills[mprn] = {}

    try:
        bill = mprn_values[reference]
    except KeyError:
        bill = mprn_values[reference] = {
            "bill_type_code": "W" if reference.startswith("CR") else "N",
            "mprn": mprn,
            "reference": reference,
            "account": account,
            "issue_date": issue_date,
            "kwh": Decimal("0"),
            "breakdown": defaultdict(int, {}),
            "net_gbp": Decimal("0.00"),
            "vat_gbp": Decimal("0.00"),
            "gross_gbp": Decimal("0.00"),
            "raw_lines": [],
            "reads": [],
        }
    if charge_period_from is not None:
        if "start_date" in bill:
            bill["start_date"] = hh_min(bill["start_date"], charge_period_from)
        else:
            bill["start_date"] = charge_period_from
    if charge_period_end is not None:
        if "finish_date" in bill:
            bill["finish_date"] = hh_max(bill["finish_date"], charge_period_end)
        else:
            bill["finish_date"] = charge_period_end
    bill["raw_lines"].append([c.value for c in sheet[row]])

    bd = bill["breakdown"]
    element_desc = get_value(sheet, row, column_map[Title.CHARGE_TYPE])
    quantity = get_dec(sheet, row, column_map[Title.QUANTITY])
    qunit = get_value(sheet, row, column_map[Title.QUNIT])
    charge = get_dec(sheet, row, column_map[Title.CHARGE]) / Decimal("100")
    total = get_dec(sheet, row, column_map[Title.TOTAL])
    if bill["bill_type_code"] == "W":
        quantity = quantity * -1
    if element_desc == "VAT":
        bill["net_gbp"] += quantity
        bill["vat_gbp"] += total
        bill["gross_gbp"] += quantity + total
        if "vat_rate" not in bd:
            bd["vat_rate"] = set()
        bd["vat_rate"].add(charge)
    else:
        element_name = ELEMENT_LOOKUP[element_desc]
        if element_name == "commodity":
            bill["kwh"] += quantity
        bd[f"{element_name}_gbp"] += total
        element_qunit = QUNIT_LOOKUP[qunit]
        bd[f"{element_name}_{element_qunit}"] += quantity
        rate_name = f"{element_name}_rate"
        if rate_name not in bd:
            bd[rate_name] = set()
        bd[rate_name].add(charge)


def _make_raw_bills(sheet):
    bills = {}
    rows = tuple(sheet.rows)
    title_row = rows[0]
    for row_index, row in enumerate(rows[1:], start=2):
        val = row[0].value
        if val not in (None, ""):
            try:
                _parse_row(bills, sheet, row_index, title_row)
            except BadRequest as e:
                raise BadRequest(f"On row {row_index}: {e.description}")

    raw_bills = []
    for mprn, mprn_values in bills.items():
        for period, raw_bill in mprn_values.items():
            raw_bills.append(raw_bill)

    return raw_bills


class Parser:
    def __init__(self, f):
        self.book = load_workbook(BytesIO(f), data_only=True)
        self.sheet = self.book.worksheets[0]

        self.last_line = None
        lines = (self._set_last_line(i, l) for i, l in enumerate(f))
        self.reader = csv.reader(lines, skipinitialspace=True)
        self._line_number = None
        self._title_line = None

    @property
    def line_number(self):
        return None if self._line_number is None else self._line_number + 1

    def _set_last_line(self, i, line):
        self._line_numer = i
        self.last_line = line
        if i == 0:
            self._title_line = line
        return line

    def make_raw_bills(self):
        return _make_raw_bills(self.sheet)
